#!/usr/bin/env python
# coding: utf-8

"""
Node layout routines.
"""

import warnings
import itertools
import numpy as np

from functools import wraps, partial
from itertools import combinations, product
from scipy.spatial import Voronoi
from scipy.spatial.distance import cdist, pdist, squareform
from scipy.optimize import minimize, NonlinearConstraint

from rpack import pack
from grandalf.graphs import Vertex, Edge, Graph
from grandalf.layouts import SugiyamaLayout, DummyVertex

from ._utils import (
    _edge_list_to_adjacency_matrix,
    _edge_list_to_adjacency_list,
    _get_subgraph,
    _get_unique_nodes,
    _get_n_points_on_a_circle,
    _invert_dict,
    _get_connected_components,
    _convert_polar_to_cartesian_coordinates,
    _get_angle,
)


DEBUG = False


def _handle_multiple_components(layout_function, packing_function):
    """Most layout algorithms only handle graphs that consist of a giant
    single component, and fail to find a suitable layout if the graph
    consists of more than component. This decorator wraps a given
    layout function such that if the graph contains more than one
    component, the layout is first computed for each individual
    component, and then the component layouts are combined using
    the packing function.

    """

    @wraps(layout_function)
    def wrapped_layout_function(edges, nodes=None, *args, **kwargs):

        # determine if there are more than one component
        adjacency_list = _edge_list_to_adjacency_list(edges, directed=False)
        components = _get_connected_components(adjacency_list)

        if nodes:
            unconnected_nodes = set(nodes) - set(_get_unique_nodes(edges))
            if unconnected_nodes:
                for node in unconnected_nodes:
                    components.append([node])

        if len(components) > 1:
            return get_layout_for_multiple_components(
                edges, components, layout_function, packing_function, *args, **kwargs)
        else:
            return layout_function(edges, *args, **kwargs)

    return wrapped_layout_function


def get_layout_for_multiple_components(edges, components,
                                       layout_function, packing_function,
                                       origin, scale, *args, **kwargs):
    """Determine suitable bounding box dimensions and placement for each
    component in the graph, and then compute the layout of each
    individual component given the constraint of the bounding box.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    components : list of sets
        The connected components of the graph.
    layout_function : function
        The function used to compute the relative positions of each node within a component.
        The arguments and key-word arguments are passed through to this function.
    packing_function : function
        The function used to arrange component layouts w.r.t. each other.
    origin : tuple
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.

    *args, **kwargs
        Passed through to layout_function

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    """

    bboxes = packing_function(components, origin, scale)

    node_positions = dict()
    for ii, (component, bbox) in enumerate(zip(components, bboxes)):
        if len(component) > 1:
            subgraph = _get_subgraph(edges, component)
            component_node_positions = layout_function(subgraph, origin=bbox[:2], scale=bbox[2:], *args, **kwargs)
            node_positions.update(component_node_positions)
        else:
            # component is a single node, which we can simply place at the centre of the bounding box
            node_positions[component.pop()] = np.array([bbox[0] + 0.5 * bbox[2], bbox[1] + 0.5 * bbox[3]])

    return node_positions


def _get_packed_component_bboxes(components, origin, scale, pad_by=0.1, power=0.8):
    """Partition the canvas given by origin and scale into bounding boxes, one for each component.

    Use rectangle packing to tightly arrange bounding boxes.

    Parameters
    ----------
    components : list of sets
        The connected components of the graph.
    origin : tuple
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    power : float, default 0.8
        The dimensions each bounding box are given by |V|^power by |V|^power,
        where |V| are the total number of nodes.

    Returns
    -------
    bboxes : list of tuples
        The (left, bottom, width height) bounding boxes for each component.

    """

    relative_dimensions = [_get_bbox_dimensions(len(component), power=power) for component in components]

    # Add a padding between boxes, such that nodes cannot end up touching in the final layout.
    # We choose a padding proportional to the dimensions of the largest box.
    maximum_width, maximum_height = np.max(relative_dimensions, axis=0)
    pad_x, pad_y = pad_by * maximum_width, pad_by * maximum_height
    padded_dimensions = [(width + pad_x, height + pad_y) for (width, height) in relative_dimensions]

    # rpack only works on integers, hence multiply by some large scalar to retain some precision;
    # NB: for some strange reason, rpack's running time is sensitive to the size of the boxes, so don't make the scalar too large
    # TODO find alternative to rpack
    scalar = 1000 / ((1 + pad_by) * max(maximum_width, maximum_height))
    integer_dimensions = [(int(scalar*width), int(scalar*height)) for width, height in padded_dimensions]
    origins = pack(integer_dimensions) # NB: rpack claims to return upper-left corners, when it actually returns lower-left corners

    bboxes = [(x, y, scalar*width, scalar*height) for (x, y), (width, height) in zip(origins, relative_dimensions)]

    # rescale boxes to canvas, effectively reversing the upscaling
    bboxes = _rescale_bboxes_to_canvas(bboxes, origin, scale)

    if DEBUG:
        import matplotlib.pyplot as plt
        from matplotlib.patches import Rectangle
        fig, ax = plt.subplots(1,1)
        for bbox in bboxes:
            ax.add_artist(Rectangle(bbox[:2], bbox[2], bbox[3], color=np.random.rand(3), zorder=-1))
        # plt.show()

    return bboxes


def _get_bbox_dimensions(n, power=0.5):
    """Given n nodes, compute appropriately scaled square bbox dimensions."""
    # TODO: factor in the dimensions of the canvas
    # such that the rescaled boxes are approximately square
    return (n**power, n**power)


def _rescale_bboxes_to_canvas(bboxes, origin, scale):
    """Convert relative bbox dimensions to absolute bbox dimensions given the dimensions of the available canvas."""
    lower_left_hand_corners = [(x, y) for (x, y, _, _) in bboxes]
    upper_right_hand_corners = [(x+w, y+h) for (x, y, w, h) in bboxes]
    minimum = np.min(lower_left_hand_corners, axis=0)
    maximum = np.max(upper_right_hand_corners, axis=0)
    total_width, total_height = maximum - minimum

    # shift to (0, 0)
    min_x, min_y = minimum
    lower_left_hand_corners = [(x - min_x, y-min_y) for (x, y) in lower_left_hand_corners]

    # rescale
    scale_x = scale[0] / total_width
    scale_y = scale[1] / total_height

    lower_left_hand_corners = [(x*scale_x, y*scale_y) for x, y in lower_left_hand_corners]
    dimensions = [(w * scale_x, h * scale_y) for (_, _, w, h) in bboxes]
    rescaled_bboxes = [(x, y, w, h) for (x, y), (w, h) in zip(lower_left_hand_corners, dimensions)]

    # shift by origin
    x0, y0 = origin
    shifted_bboxes = [(x + x0, y + y0, w, h) for (x, y, w, h) in rescaled_bboxes]
    return shifted_bboxes


def _get_side_by_side_component_bboxes(components, origin, scale, pad_by=0.05):
    """Partition the canvas given by origin and scale into bounding boxes, one for each component.

    Position bounding boxes next to each other.

    Parameters
    ----------
    components : list of sets
        The connected components of the graph.
    origin : tuple
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.

    Returns
    -------
    bboxes : list of tuples
        The (left, bottom, width height) bounding boxes for each component.

    """

    relative_dimensions = [(len(component), 1) for component in components]

    # Add a padding between boxes, such that nodes cannot end up touching in the final layout.
    # We choose a padding proportional to the dimensions of the largest box.
    maximum_dimensions = np.max(relative_dimensions, axis=0)
    pad_x, pad_y = pad_by * maximum_dimensions
    padded_dimensions = [(width + pad_x, height + pad_y) for (width, height) in relative_dimensions]

    x = 0
    origins = []
    for (dx, _) in padded_dimensions:
        origins.append((x, 0))
        x += dx

    bboxes = [(x, y, width, height) for (x, y), (width, height) in zip(origins, relative_dimensions)]

    # rescale boxes to canvas, effectively reversing the upscaling
    bboxes = _rescale_bboxes_to_canvas(bboxes, origin, scale)

    if DEBUG:
        import matplotlib.pyplot as plt
        from matplotlib.patches import Rectangle
        fig, ax = plt.subplots(1,1)
        for bbox in bboxes:
            ax.add_artist(Rectangle(bbox[:2], bbox[2], bbox[3], color=np.random.rand(3)))
        plt.show()

    return bboxes


_rectangle_pack_multiple_components = \
    partial(_handle_multiple_components, packing_function=_get_packed_component_bboxes)

_arrange_multiple_components_side_by_side = \
    partial(_handle_multiple_components, packing_function=_get_side_by_side_component_bboxes)


def _get_fr_repulsion(distance, direction, k):
    """Compute repulsive forces."""
    with np.errstate(divide='ignore', invalid='ignore'):
        magnitude = k**2 / distance
    vectors = direction * magnitude[..., None]
    # Note that we cannot apply the usual strategy of summing the array
    # along either axis and subtracting the trace,
    # as the diagonal of `direction` is np.nan, and any sum or difference of
    # NaNs is just another NaN.
    # Also we do not want to ignore NaNs by using np.nansum, as then we would
    # potentially mask the existence of off-diagonal zero distances.
    for ii in range(vectors.shape[-1]):
        np.fill_diagonal(vectors[:, :, ii], 0)
    return np.sum(vectors, axis=0)


def _get_fr_attraction(distance, direction, adjacency, k):
    """Compute attractive forces."""
    magnitude = 1./k * distance**2 * adjacency
    vectors = -direction * magnitude[..., None] # NB: the minus!
    for ii in range(vectors.shape[-1]):
        np.fill_diagonal(vectors[:, :, ii], 0)
    return np.sum(vectors, axis=0)


@_rectangle_pack_multiple_components
def get_fruchterman_reingold_layout(edges,
                                    edge_weights        = None,
                                    k                   = None,
                                    origin              = (0, 0),
                                    scale               = (1, 1),
                                    pad_by              = 0.05,
                                    initial_temperature = 1.,
                                    total_iterations    = 50,
                                    node_size           = 0,
                                    node_positions      = None,
                                    fixed_nodes         = None,
                                    validate_positions  = True,
                                    get_repulsion       = _get_fr_repulsion,
                                    get_attraction      = _get_fr_attraction,
                                    *args, **kwargs):
    """'Spring' or Fruchterman-Reingold node layout.

    Uses the Fruchterman-Reingold algorithm [Fruchterman1991]_ to compute node positions.
    This algorithm simulates the graph as a physical system, in which nodes repell each other.
    For connected nodes, this repulsion is counteracted by an attractive force exerted by the edges, which are simulated as springs.
    The resulting layout is hence often referred to as a 'spring' layout.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    edge_weights : dict
        Mapping of edges to edge weights.
    k : float or None, default None
        Expected mean edge length. If None, initialized to the sqrt(area / total nodes).
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code-block::

           xmin = origin[0] + pad_by * scale[0]
           ymin = origin[1] + pad_by * scale[1]
           xmax = origin[0] + scale[0] - pad_by * scale[0]
           ymax = origin[1] + scale[1] - pad_by * scale[1]

    total_iterations : int, default 50
        Number of iterations.
    initial_temperature: float, default 1.
        Temperature controls the maximum node displacement on each iteration.
        Temperature is decreased on each iteration to eventually force the algorithm into a particular solution.
        The size of the initial temperature determines how quickly that happens.
        Values should be much smaller than the values of `scale`.
    node_size : scalar or dict, default 0.
        Size (radius) of nodes.
        Providing the correct node size minimises the overlap of nodes in the graph,
        which can otherwise occur if there are many nodes, or if the nodes differ considerably in size.
    node_positions : dict or None, default None
        Mapping of nodes to their (initial) x,y positions. If None are given,
        nodes are initially placed randomly within the bounding box defined by `origin` and `scale`.
        If the graph has multiple components, explicit initial positions may result in a ValueError,
        if the initial positions fall outside of the area allocated to that specific component.
    fixed_nodes : list or None, default None
        Nodes to keep fixed at their initial positions.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Fruchterman1991] Fruchterman, TMJ and Reingold, EM (1991) ‘Graph drawing by force‐directed placement’,
       Software: Practice and Experience

    """

    assert len(edges) > 0, "The list of edges has to be non-empty."

    # This is just a wrapper around `_fruchterman_reingold`, which implements (the loop body of) the algorithm proper.
    # This wrapper handles the initialization of variables to their defaults (if not explicitely provided),
    # and checks inputs for self-consistency.

    origin = np.array(origin)
    scale = np.array(scale)
    assert len(origin) == len(scale), \
        "Arguments `origin` (d={}) and `scale` (d={}) need to have the same number of dimensions!".format(len(origin), len(scale))
    dimensionality = len(origin)

    if fixed_nodes is None:
        fixed_nodes = []

    connected_nodes = _get_unique_nodes(edges)

    if node_positions is None: # assign random starting positions to all nodes
        node_positions_as_array = np.random.rand(len(connected_nodes), dimensionality) * scale + origin
        unique_nodes = connected_nodes

    else:

        if validate_positions:
            # 1) check input dimensionality
            dimensionality_node_positions = np.array(list(node_positions.values())).shape[1]
            assert dimensionality_node_positions == dimensionality, \
                "The dimensionality of values of `node_positions` (d={}) must match the dimensionality of `origin`/ `scale` (d={})!".format(dimensionality_node_positions, dimensionality)

            is_valid = _is_within_bbox(list(node_positions.values()), origin=origin, scale=scale)
            if not np.all(is_valid):
                error_message = "Some given node positions are not within the data range specified by `origin` and `scale`!"
                error_message += "\n\tOrigin : {}, {}".format(*origin)
                error_message += "\n\tScale  : {}, {}".format(*scale)
                error_message += "\nThe following nodes do not fall within this range:"
                for ii, (node, position) in enumerate(node_positions.items()):
                    if not is_valid[ii]:
                        error_message += "\n\t{} : {}".format(node, position)
                error_message += "\nThis error can occur if the graph contains multiple components but some or all node positions are initialised explicitly (i.e. node_positions != None)."
                raise ValueError(error_message)

        # 2) handle discrepancies in nodes listed in node_positions and nodes extracted from edges
        if set(node_positions.keys()) == set(connected_nodes):
            # all starting positions are given;
            # no superfluous nodes in node_positions;
            # nothing left to do
            unique_nodes = connected_nodes
        else:
            # some node positions are provided, but not all
            for node in connected_nodes:
                if not (node in node_positions):
                    warnings.warn("Position of node {} not provided. Initializing to random position within frame.".format(node))
                    node_positions[node] = np.random.rand(2) * scale + origin

            unconnected_nodes = []
            for node in node_positions:
                if not (node in connected_nodes):
                    unconnected_nodes.append(node)
                    fixed_nodes.append(node)
                    # warnings.warn("Node {} appears to be unconnected. The current node position will be kept.".format(node))

            unique_nodes = connected_nodes + unconnected_nodes

        node_positions_as_array = np.array([node_positions[node] for node in unique_nodes])

    total_nodes = len(unique_nodes)

    if isinstance(node_size, (int, float)):
        node_size = node_size * np.ones((total_nodes))
    elif isinstance(node_size, dict):
        node_size = np.array([node_size[node] if node in node_size else 0. for node in unique_nodes])

    adjacency = _edge_list_to_adjacency_matrix(
        edges, edge_weights=edge_weights, unique_nodes=unique_nodes)

    # Forces in FR are symmetric.
    # Hence we need to ensure that the adjacency matrix is also symmetric.
    adjacency = adjacency + adjacency.transpose()

    if fixed_nodes:
        is_mobile = np.array([False if node in fixed_nodes else True for node in unique_nodes], dtype=bool)

        mobile_positions = node_positions_as_array[is_mobile]
        fixed_positions = node_positions_as_array[~is_mobile]

        mobile_node_sizes = node_size[is_mobile]
        fixed_node_sizes = node_size[~is_mobile]

        # reorder adjacency
        total_mobile = np.sum(is_mobile)
        reordered = np.zeros((adjacency.shape[0], total_mobile))
        reordered[:total_mobile, :total_mobile] = adjacency[is_mobile][:, is_mobile]
        reordered[total_mobile:, :total_mobile] = adjacency[~is_mobile][:, is_mobile]
        adjacency = reordered
    else:
        is_mobile = np.ones((total_nodes), dtype=bool)

        mobile_positions = node_positions_as_array
        fixed_positions = np.zeros((0, 2))

        mobile_node_sizes = node_size
        fixed_node_sizes = np.array([])

    if k is None:
        area = np.prod(scale)
        k = np.sqrt(area / float(total_nodes))

    temperatures = _get_temperature_decay(initial_temperature, total_iterations)

    # --------------------------------------------------------------------------------
    # main loop

    for ii, temperature in enumerate(temperatures):
        candidate_positions = _fruchterman_reingold(mobile_positions, fixed_positions,
                                                    mobile_node_sizes, fixed_node_sizes,
                                                    adjacency, temperature, k,
                                                    get_repulsion, get_attraction)
        is_valid = _is_within_bbox(candidate_positions, origin=origin, scale=scale)
        mobile_positions[is_valid] = candidate_positions[is_valid]

    # --------------------------------------------------------------------------------
    # format output

    node_positions_as_array[is_mobile] = mobile_positions

    if np.all(is_mobile):
        node_positions_as_array = _fit_to_frame(node_positions_as_array, origin, scale, pad_by)

    node_positions = dict(zip(unique_nodes, node_positions_as_array))

    return node_positions


def _is_within_bbox(points, origin, scale):
    """Check if each of the given points is within the bounding box given by origin and scale."""
    minima = np.array(origin)
    maxima = minima + np.array(scale)
    return np.all(np.logical_and(points >= minima, points <= maxima), axis=1)


def _get_temperature_decay(initial_temperature, total_iterations, mode='quadratic', eps=1e-9):
    """Compute all temperature values for a given initial temperature and decay model."""
    x = np.linspace(0., 1., total_iterations)
    if mode == 'quadratic':
        y = (x - 1.)**2 + eps
    elif mode == 'linear':
        y = (1. - x) + eps
    else:
        raise ValueError("Argument `mode` one of: 'linear', 'quadratic'.")
    return initial_temperature * y


def _fruchterman_reingold(mobile_positions, fixed_positions,
                          mobile_node_radii, fixed_node_radii,
                          adjacency, temperature, k,
                          get_repulsion, get_attraction):
    """Inner loop of Fruchterman-Reingold layout algorithm."""

    combined_positions = np.concatenate([mobile_positions, fixed_positions], axis=0)
    combined_node_radii = np.concatenate([mobile_node_radii, fixed_node_radii])

    delta = mobile_positions[np.newaxis, :, :] - combined_positions[:, np.newaxis, :]
    distance = np.linalg.norm(delta, axis=-1)

    # alternatively: (hack adapted from igraph)
    if np.sum(distance==0) - np.trace(distance==0) > 0: # i.e. if off-diagonal entries in distance are zero
        warnings.warn("Some nodes have the same position; repulsion between the nodes is undefined.")
        rand_delta = np.random.rand(*delta.shape) * 1e-9
        is_zero = distance <= 0
        delta[is_zero] = rand_delta[is_zero]
        distance = np.linalg.norm(delta, axis=-1)

    # subtract node radii from distances to prevent nodes from overlapping
    distance -= mobile_node_radii[np.newaxis, :] + combined_node_radii[:, np.newaxis]

    # prevent distances from becoming less than zero due to overlap of nodes
    distance[distance <= 0.] = 1e-6 # 1e-13 is numerical accuracy, and we will be taking the square shortly

    with np.errstate(divide='ignore', invalid='ignore'):
        direction = delta / distance[..., None] # i.e. the unit vector

    # calculate forces
    repulsion    = get_repulsion(distance, direction, k)
    attraction   = get_attraction(distance, direction, adjacency, k)
    displacement = attraction + repulsion

    # limit maximum displacement using temperature
    displacement_length = np.linalg.norm(displacement, axis=-1)
    displacement = displacement / displacement_length[:, None] * np.clip(displacement_length, None, temperature)[:, None]

    mobile_positions = mobile_positions + displacement

    return mobile_positions


def _rescale_to_frame(node_positions, origin, scale):
    """Rescale node positions such that all nodes are within the bounding box."""
    node_positions = node_positions.copy() # force copy, as otherwise the `fixed_nodes` argument is effectively ignored
    node_positions -= np.min(node_positions, axis=0)
    # normalize only when nodes are not already aligned, otherwise we divide by zero
    np.divide(node_positions, np.ptp(node_positions, axis=0), where=np.ptp(node_positions, axis=0) != 0, out=node_positions)
    # if nodes are aligned in any one dimension, place in the middle along that dimension
    np.add(node_positions, np.full(node_positions.ndim, 0.5), where=np.ptp(node_positions, axis=0) == 0, out=node_positions)
    node_positions *= scale[None, ...]
    node_positions += origin[None, ...]
    return node_positions


@_rectangle_pack_multiple_components
def get_random_layout(edges, origin=(0,0), scale=(1,1)):
    nodes = _get_unique_nodes(edges)
    return {node : np.random.rand(2) * scale + origin for node in nodes}


@_rectangle_pack_multiple_components
def get_sugiyama_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_size=3, total_iterations=3):
    """'Dot' or Sugiyama node layout.

    Uses the Sugiyama algorithm [Sugiyama1981]_ to compute node positions.
    This function is a wrapper around the SugiyamaLayout class in grandalf.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code-block::

           xmin = origin[0] + pad_by * scale[0]
           ymin = origin[1] + pad_by * scale[1]
           xmax = origin[0] + scale[0] - pad_by * scale[0]
           ymax = origin[1] + scale[1] - pad_by * scale[1]

    total_iterations : int, default 50
        Number of iterations.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Sugiyama1981] Sugiyama, K; Tagawa, S; Toda, M (1981) 'Methods for visual understanding of hierarchical system structures',
           IEEE Transactions on Systems, Man, and Cybernetics

    """

    # TODO potentially test that graph is a DAG
    nodes = _get_unique_nodes(edges)
    graph = _get_grandalf_graph(edges, nodes, node_size)

    layout = SugiyamaLayout(graph.C[0])
    layout.init_all()
    layout.draw(total_iterations)

    # extract node positions
    node_positions = dict()
    for layer in layout.layers:
        for vertex in layer:
            if not isinstance(vertex, DummyVertex):
                # The DummyVertex class is used by the sugiyama layout to represent
                # *long* edges, i.e. edges that span over several ranks.
                # For these edges, a DummyVertex is inserted in every inner layer.
                # Here we ignore them, as they are not part of the final layout.
                node_positions[vertex.data] = vertex.view.xy

    # rescale to canvas
    # TODO: by rescaling, we effectively ignore the node_size argument
    nodes, positions = zip(*node_positions.items())
    positions = _rescale_to_frame(np.array(positions), np.array(origin) + pad_by * np.array(scale), (1 - pad_by) * np.array(scale))

    # place roots on top, leaves on bottom
    positions[:, 1] -= origin[1] + scale[1]
    positions[:, 1] *= -1

    return dict(zip(nodes, positions))


def _get_grandalf_graph(edges, nodes, node_size):
    """Construct a grandalf.Graph object from the given edge list, node list, and node sizes."""
    node_to_grandalf_vertex = dict()
    for node in nodes:
        # initialize vertex object
        vertex = Vertex(node)

        # initialize view
        if isinstance(node_size, (int, float)):
            vertex.view = vertex_view(2 * node_size, 2 * node_size)
        elif isinstance(node_size, dict):
            vertex.view = vertex_view(2 * node_size[node], 2 * node_size[node])
        else:
            raise TypeError

        node_to_grandalf_vertex[node] = vertex

    V = list(node_to_grandalf_vertex.values())
    E = [Edge(node_to_grandalf_vertex[source], node_to_grandalf_vertex[target]) for source, target in edges]
    G = Graph(V, E)
    return G


class vertex_view(object):
    def __init__(self, w, h):
        self.w = w
        self.h = h


@_rectangle_pack_multiple_components
def get_radial_tree_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_size=3, total_iterations=3):
    """Radial tree layout.

    Uses the Sugiyama algorithm [Sugiyama1981]_ to compute node positions.
    This function is a wrapper around the SugiyamaLayout class in grandalf.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code-block::

           xmin = origin[0] + pad_by * scale[0]
           ymin = origin[1] + pad_by * scale[1]
           xmax = origin[0] + scale[0] - pad_by * scale[0]
           ymax = origin[1] + scale[1] - pad_by * scale[1]

    total_iterations : int, default 50
        Number of iterations.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Sugiyama1981] Sugiyama, K; Tagawa, S; Toda, M (1981) 'Methods for visual understanding of hierarchical system structures',
           IEEE Transactions on Systems, Man, and Cybernetics

    """

    sugiyama_positions = get_sugiyama_layout(edges,
                                             origin           = (0, 0),
                                             scale            = (1, 1),
                                             pad_by           = 0,
                                             node_size        = node_size,
                                             total_iterations = total_iterations
    )

    # determine the size of the largest layer
    nodes_per_layer = dict()
    for node, (_, y) in sugiyama_positions.items():
        if y in nodes_per_layer:
            nodes_per_layer[y] += 1
        else:
            nodes_per_layer[y] = 1
    max_nodes_per_layer = np.max(list(nodes_per_layer.values()))

    # alternatively:
    # y_coordinates = np.array(list(sugiyama_positions.values()))[:, 1]
    # (_, max_nodes_per_layer), = Counter(y_coordinates).most_common(1)

    max_angle = 2 * np.pi * max_nodes_per_layer / (max_nodes_per_layer + 1)
    max_radius = np.min(scale) / 2
    max_radius *= (1 - pad_by) # shrink to make room for node artists, labels, and annotations
    offset = np.array([origin[0] + 0.5 * scale[0], origin[1] + 0.5 * scale[1]])

    node_positions = dict()
    for node, (x, y) in sugiyama_positions.items():
        angle = max_angle * x
        radius = max_radius * (1 - y)
        node_positions[node] = _convert_polar_to_cartesian_coordinates(radius, angle) + offset

    return node_positions


@_rectangle_pack_multiple_components
def get_circular_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_order=None, reduce_edge_crossings=True):
    """Circular node layout.

    By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are minimised.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code-block::

           xmin = origin[0] + pad_by * scale[0]
           ymin = origin[1] + pad_by * scale[1]
           xmax = origin[0] + scale[0] - pad_by * scale[0]
           ymax = origin[1] + scale[1] - pad_by * scale[1]

    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        Implies :code:`reduce_edge_crossings` is :code:`False`.
    reduce_edge_crossings : bool, default True
        If True, attempts to minimize edge crossings via the algorithm outlined in [Baur2005]_.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    See also
    --------
    :py:func:`_get_preorderd_circular_layout`

    References
    ----------
    .. [Baur2005] Baur & Brandes (2005) Crossing reduction in circular layouts.

    """
    nodes = _get_unique_nodes(edges)

    if node_order:
        nodes = [node for node in node_order if node in nodes]
    elif reduce_edge_crossings:
        if not _is_complete(edges):
            # remove self-loops as these should have no bearing on the solution
            edges = [(source, target) for source, target in edges if source != target]
            nodes = _reduce_crossings(edges)
    else:
        try: # emulate networkx behavior
            nodes = sorted(nodes)
        except TypeError: # probably due to mixed node types
            pass

    return _get_preordered_circular_layout(nodes, origin, scale, pad_by)


def _get_preordered_circular_layout(node_order, origin, scale, pad_by=0.05):
    center = np.array(origin) + 0.5 * np.array(scale)
    radius = np.min(scale) / 2
    radius *= (1 - pad_by) # fudge factor to make space for self-loops, annotations, etc
    positions = _get_n_points_on_a_circle(center, radius, len(node_order), start_angle=0)
    return dict(zip(node_order, positions))


def _is_complete(edges):
    """Check if the graph is fully connected."""
    nodes = _get_unique_nodes(edges)
    minimal_complete_graph = list(combinations(nodes, 2))

    if len(edges) < len(minimal_complete_graph):
        return False

    for edge in minimal_complete_graph:
        if (edge not in edges) and (edge[::-1] not in edges):
            return False

    return True


def _reduce_crossings(edges):
    """Implements Baur & Brandes (2005) Crossing reduction in circular layouts."""
    adjacency_list = _edge_list_to_adjacency_list(edges, directed=False)
    node_order = _initialize_node_order(adjacency_list)
    node_order = _optimize_node_order(adjacency_list, node_order)
    return node_order


def _initialize_node_order(node_to_neighbours):
    """Implements "Connectivity & Crossings" variant from Baur & Brandes (2005)."""
    nodes = list(node_to_neighbours.keys())
    ordered = nodes[:1]
    remaining = nodes[1:]
    closed_edges = []
    start, = ordered
    open_edges = set([(start, neighbour) for neighbour in node_to_neighbours[start]])

    while remaining:
        minimum_unplaced_neighbours = np.inf
        maximum_placed_neighbours = 0
        for ii, node in enumerate(remaining):
            placed_neighbours = len([neighbour for neighbour in node_to_neighbours[node] if neighbour in ordered])
            unplaced_neighbours = len([neighbour for neighbour in node_to_neighbours[node] if neighbour not in ordered])
            if (placed_neighbours > maximum_placed_neighbours) or \
               ((placed_neighbours == maximum_placed_neighbours) and (unplaced_neighbours < minimum_unplaced_neighbours)):
                maximum_placed_neighbours = placed_neighbours
                minimum_unplaced_neighbours = unplaced_neighbours
                selected = node
                selected_idx = ii

        remaining.pop(selected_idx)
        closed_edges = set([(node, selected) for node in node_to_neighbours[selected] if node in ordered])
        open_edges -= closed_edges
        a = _get_total_crossings(ordered + [selected] + remaining, closed_edges, open_edges)
        b = _get_total_crossings([selected] + ordered + remaining, closed_edges, open_edges)
        if a < b:
            ordered.append(selected)
        else:
            ordered.insert(0, selected)
        open_edges |= set([(selected, node) for node in node_to_neighbours[selected] if node not in ordered])

    return ordered


def _get_total_crossings(node_order, edges1, edges2=None):
    """Compute the number of crossings for a given node order."""
    if edges2 is None:
        edges2 = edges1
    total_crossings = 0
    for edge1, edge2 in itertools.product(edges1, edges2):
        total_crossings += int(_is_cross(node_order, edge1, edge2))
    return total_crossings


def _is_cross(node_order, edge_1, edge_2):
    """Check if two edges cross, given the node order."""
    (s1, t1), = np.where(np.in1d(node_order, edge_1, assume_unique=True))
    (s2, t2), = np.where(np.in1d(node_order, edge_2, assume_unique=True))

    if (s1 < s2 < t1 < t2) or (s2 < s1 < t2 < t2):
        return True
    else:
        return False


def _optimize_node_order(node_to_neighbours, node_order=None, max_iterations=100):
    """Implement circular sifting as outlined in Baur & Brandes (2005)."""
    if node_order is None:
        node_order = list(node_to_neighbours.keys())
    node_order = np.array(node_order)

    node_to_edges = dict()
    for node, neighbours in node_to_neighbours.items():
        node_to_edges[node] = [(node, neighbour) for neighbour in neighbours]

    undirected_edges = set()
    for edges in node_to_edges.values():
        for edge in edges:
            if edge[::-1] not in undirected_edges:
                undirected_edges.add(edge)
    total_crossings = _get_total_crossings(node_order, undirected_edges)

    total_nodes = len(node_order)
    for iteration in range(max_iterations):
        previous_best = total_crossings
        for u in node_order.copy():
            best_order = node_order.copy()
            minimum_crossings = total_crossings
            (ii,), = np.where(node_order == u)
            for jj in range(ii+1, ii+total_nodes):
                v = node_order[jj%total_nodes]
                cuv = _get_total_crossings(node_order, node_to_edges[u], node_to_edges[v])
                node_order = _swap_values(node_order, u, v)
                cvu = _get_total_crossings(node_order, node_to_edges[u], node_to_edges[v])
                total_crossings = total_crossings - cuv + cvu
                if total_crossings < minimum_crossings:
                    best_order = node_order.copy()
                    minimum_crossings = total_crossings
            node_order = best_order
            total_crossings = minimum_crossings

        improvement = previous_best - total_crossings
        if not improvement:
            break

    if (iteration + 1) == max_iterations:
        warnings.warn("Maximum number of iterations reached. Aborting further node layout optimisations.")

    return node_order


def _swap_values(arr, value_1, value_2):
    """Swap all occurrences of two values in a given array."""
    arr = arr.copy()
    is_value_1 = arr == value_1
    is_value_2 = arr == value_2
    arr[is_value_1] = value_2
    arr[is_value_2] = value_1
    return arr


def _reduce_node_overlap(node_positions, origin, scale, fixed_nodes=None, eta=0.1, total_iterations=10):
    """Use a constrained version of Lloyd's algorithm to move nodes apart from each other.

    References
    ----------
    https://en.wikipedia.org/wiki/Lloyd%27s_algorithm
    """

    unique_nodes = list(node_positions.keys())
    positions = np.array(list(node_positions.values()))

    if fixed_nodes:
        is_mobile = np.array([False if node in fixed_nodes else True for node in unique_nodes], dtype=bool)
    else:
        is_mobile = np.ones((len(unique_nodes)), dtype=bool)

    for _ in range(total_iterations):
        centroids = _get_voronoi_centroids(positions)
        delta = centroids - positions
        new = positions + eta * delta
        # constrain Lloyd's algorithm by only updating positions where the new position is within the bbox
        valid = _is_within_bbox(new, origin, scale)
        mask = np.logical_and(valid, is_mobile)
        positions[mask] = new[mask]

    return dict(zip(unique_nodes, positions))


def _remove_node_overlap(node_positions, node_size, origin, scale, fixed_nodes=None, tolerance=1e-6, maximum_iterations=100):
    """Uses a constrained variation of Lloyd's algorithm to move nodes apart from each other until none overlap."

    References
    ----------
    https://en.wikipedia.org/wiki/Lloyd%27s_algorithm

    """

    unique_nodes = list(node_positions.keys())
    positions = np.array(list(node_positions.values()))
    radii = np.array([node_size[node] for node in node_positions])

    if fixed_nodes:
        is_mobile = np.array([False if node in fixed_nodes else True for node in unique_nodes], dtype=bool)
    else:
        is_mobile = np.ones((len(unique_nodes)), dtype=bool)

    minimum_distances = radii[np.newaxis, :] + radii[:, np.newaxis]
    minimum_distances[np.diag_indices_from(minimum_distances)] = 0 # ignore distances to self

    # Initialize the first loop.
    distances = cdist(positions, positions)
    displacements = np.max(np.clip(minimum_distances - distances, 0, None), axis=-1)

    ctr = 0
    while np.any(displacements > tolerance) & (ctr < maximum_iterations):
        centroids = _get_voronoi_centroids(positions)

        # Compute the direction from each point towards its corresponding Voronoi centroid.
        deltas = centroids - positions
        magnitudes = np.linalg.norm(deltas, axis=-1)
        directions = deltas / magnitudes[:, np.newaxis]

        # Mask NaNs that arise if the magnitude is zero, i.e. the point is already center of the Voronoi cell.
        directions[np.isnan(directions)] = 0

        # Step into the direction of the centroid.
        # Clipping prevents overshooting of the centroid when stepping into the direction of the centroid.
        # We step by half the displacement as the other overlapping point will be moved in approximately the opposite direction.
        new = positions + np.clip(0.5 * displacements, None, magnitudes)[:, np.newaxis] * directions

        # Constrain Lloyd's algorithm by only updating positions where the new position is within the bbox.
        valid = _is_within_bbox(new, origin, scale)
        mask = np.logical_and(valid, is_mobile)
        positions[mask] = new[mask]

        # Initialize next loop.
        distances = cdist(positions, positions)
        displacements = np.max(np.clip(minimum_distances - distances, 0, None), axis=-1)
        ctr += 1

    return dict(zip(unique_nodes, positions))


def _get_voronoi_centroids(positions):
    """Construct a Voronoi diagram from the given node positions and determine the center of each cell."""
    voronoi = Voronoi(positions)
    centroids = np.zeros_like(positions)
    for ii, idx in enumerate(voronoi.point_region):
        region = [jj for jj in voronoi.regions[idx] if jj != -1] # i.e. ignore points at infinity; TODO: compute correctly clipped regions
        centroids[ii] = _get_centroid(voronoi.vertices[region])
    return centroids


def _clip_to_frame(positions, origin, scale):
    """Prevent node positions from leaving the bounding box given by origin and scale."""
    origin = np.array(origin)
    scale = np.array(scale)
    for ii, (minimum, maximum) in enumerate(zip(origin, origin+scale)):
        positions[:, ii] = np.clip(positions[:, ii], minimum, maximum)
    return positions


def _get_centroid(polygon):
    """Compute the centroid of a polygon."""
    # TODO: formula may be incorrect; correct one here:
    # https://en.wikipedia.org/wiki/Centroid#Of_a_polygon
    return np.mean(polygon, axis=0)


@_arrange_multiple_components_side_by_side
def get_linear_layout(edges, origin=(0, 0), scale=(1, 1), pad_by=0.05, node_order=None, reduce_edge_crossings=True):
    """Linear node layout.

    If :code:`reduce_edge_crossings` is set to :code:`True`, the algorithm
    attempts to minimize edge crossings via the algorithm outlined in [Baur2005]_.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        Implies :code:`reduce_edge_crossings` is :code:`False`.
    reduce_edge_crossings : bool, default True
        If True, attempts to minimize edge crossings via the algorithm outlined in [Baur2005]_.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    See also
    --------
    _get_preordered_linear_layout

    References
    ----------
    .. [Baur2005] Baur & Brandes (2005) Crossing reduction in circular layouts.

    """

    nodes = _get_unique_nodes(edges)

    if node_order:
        nodes = [node for node in node_order if node in nodes]
    elif reduce_edge_crossings:
        if not _is_complete(edges):
            # remove self-loops as these should have no bearing on the solution
            edges = [(source, target) for source, target in edges if source != target]
            nodes = _reduce_crossings(edges)
            # The algorithm in _reduce_crossings assumes that nodes are arranged in a circle.
            # As a consequence, when nodes are arranged along a line, a community may be split
            # with members on both ends of the line.
            # Here we find a better split by minimising the total edge length.
            nodes = _minimize_total_edge_length(nodes, edges)
    else:
        try:
            nodes = sorted(nodes)
        except TypeError: # probably due to mixed node types
            pass

    return _get_preordered_linear_layout(nodes, origin, scale, pad_by)


def _get_preordered_linear_layout(node_order, origin, scale, pad_by=0.05):
    total_nodes = len(node_order)
    x = np.linspace(origin[0] + pad_by * scale[0], origin[0] + (1 - pad_by) * scale[0], total_nodes)
    y = np.full(total_nodes, origin[1] + 0.5 * scale[1])
    return dict(zip(node_order, np.c_[x, y]))


def _minimize_total_edge_length(nodes, edges):
    total_iterations = len(nodes)
    output = nodes
    optimum = np.inf
    for ii in range(total_iterations):
        node_to_position = dict(zip(nodes, range(len(nodes))))
        total_edge_length = np.sum([np.abs(node_to_position[source] - node_to_position[target]) for (source, target) in edges])
        if total_edge_length < optimum:
            output = nodes
            optimum = total_edge_length
        nodes = np.r_[nodes[1:], nodes[0]]
    return output


def get_bipartite_layout(edges, nodes=None, subsets=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True):
    """Bipartite node layout.

    By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are reduced.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    subsets : list
        The two layers of the graph. If None, a two-coloring is used to separate the nodes into two subsets.
        However, if the graph consists of multiple components, this partitioning into two layers is ambiguous, as multiple solutions exist.
    origin : tuple
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    reduce_edge_crossings : bool, default True
        If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Eades1994] Eades & Wormald (1994) Edge crossings in drawings of bipartite graphs.

    """

    adjacency_list = _edge_list_to_adjacency_list(edges, directed=False)

    if nodes:
        for node in nodes:
            adjacency_list.setdefault(node, set())

    if subsets:
        left, right = subsets
    else:
        if len(_get_connected_components(adjacency_list)) > 1:
            import warnings
            msg = "The graph consistst of multiple components, and hence the partitioning into two subsets/layers is ambiguous!"
            msg += "\n"
            msg += "Use the `subsets` argument to explicitly specify the desired partitioning."
            warnings.warn(msg)
        left, right = _get_bipartite_sets(adjacency_list)

    if reduce_edge_crossings:
        if not _is_complete_bipartite(edges, left, right):
            left, right = _reduce_crossings_bipartite(adjacency_list, left, right)

    # shrink frame to apply padding
    origin = np.array(origin) + pad_by * np.array(scale)
    scale = np.array(scale) * (1 - 2 * pad_by)

    # determine the spacing between nodes within a subset
    if len(edges) == 1:
        spacing = 1.
    elif len(left) > len(right):
        spacing = scale[1] / (len(left) - 1)
    else:
        spacing = scale[1] / (len(right) - 1)

    # set node positions
    node_positions = dict()
    for subset, xx in ((left, origin[0]), (right, origin[0] + scale[0])):
        y = spacing * np.arange(len(subset))
        y -= np.mean(y)
        y += scale[1] / 2
        for node, yy in zip(subset, y):
            node_positions[node] = (xx, yy)

    return node_positions


def _get_bipartite_sets(adjacency_list):
    colour = _get_two_colouring(adjacency_list)
    left = [node for node in colour if colour[node] == 0]
    right = [node for node in colour if colour[node] == 1]
    return left, right


def _get_two_colouring(adjacency_list):
    # Adapted from networkx.algorithms.bipartite.color.
    colour = dict()
    for node in adjacency_list:
        if (node in colour):
            continue
        elif len(adjacency_list[node]) == 0:
            colour[node] = 0
        else:
            queue = [node]
            colour[node] = 1
            while queue:
                node = queue.pop()
                for neighbour in adjacency_list[node]:
                    if neighbour in colour:
                        if colour[neighbour] == colour[node]:
                            raise Exception("Graph is not bipartite.")
                    else:
                        colour[neighbour] = 1 - colour[node]
                        queue.append(neighbour)
    return colour


def _is_complete_bipartite(edges, left, right):
    """Check if the bipartite graph is fully connected."""
    minimal_complete_graph = list(product(left, right))

    if len(edges) < len(minimal_complete_graph):
        return False

    for edge in minimal_complete_graph:
        if (edge not in edges) and (edge[::-1] not in edges):
            return False

    return True


def _reduce_crossings_bipartite(adjacency_list, left, right):
    """Reduce the number of crossings in a bipartite graph using the median heuristic proposed in Eades & Wormald (1994)."""

    left_ranks = {node : ii for ii, node in enumerate(left)}
    right_ranks = dict()
    for node in right:
        neighbours = adjacency_list[node]
        if neighbours:
            right_ranks[node] = np.median([left_ranks[neighbour] for neighbour in neighbours])
        else:
            right_ranks[node] = 0

    # TODO: break ties.
    # If one node has an even number of neighbours and the other an odd number, than the odd one should have the lower rank.
    # For the other two cases (even/even, odd/odd), no tie-break procedure seems specified in the paper.

    return left, sorted(right_ranks, key=right_ranks.get)


def get_multipartite_layout(edges, layers, layer_positions=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True, uniform_node_spacing=True):
    """Layered node layout for a multipartite graph.

    By default, this implementation uses a heuristic to arrange the nodes such that the edge crossings are reduced.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    layers : list
        List of node subsets, one for each layer of the graph.
    layer_positions : list, default None
        A list of x-coordinates, one for each layer.
        If None provided, layers are placed evenly between origin[0] and origin[0] + scale[0].
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    reduce_edge_crossings : bool, default True
        If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.
    uniform_node_spacing : bool, default True
        If True, the spacing between nodes is uniform across layers.
        Otherwise, nodes in each layer are distributed evenly across the full height of the canvas.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Eades1994] Eades & Wormald (1994) Edge crossings in drawings of bipartite graphs.

    """

    # shrink frame to apply padding
    origin = np.array(origin) + pad_by * np.array(scale)
    scale = np.array(scale) * (1 - 2 * pad_by)

    # set the space between nodes
    if uniform_node_spacing:
        try:
            spacing = scale[1] / (np.max([len(layer) for layer in layers]) - 1)
            node_spacings =  [spacing * np.ones_like(layer, dtype=float) for layer in layers]
        except ZeroDivisionError:
            # The graph has at most a single edge between each pair of layers.
            spacing = 1
            node_spacings = [spacing * np.ones_like(layer, dtype=float) for layer in layers]
    else:
        node_spacings = []
        for ii, layer in enumerate(layers):
            if len(layer) > 1:
                node_spacings.append(1./(len(layer) - 1) * np.ones_like(layer, float))
            else:
                node_spacings.append(np.ones_like(layer, float))

    # set the space between layers
    if layer_positions is None:
        layer_positions = np.linspace(origin[0], origin[0] + scale[0], len(layers))
    else:
        if (np.min(layer_positions) < origin[0]) or (np.max(layer_positions) > origin[0] + scale[0]):
            import warnings
            warnings.warn("Some layer positions are outside the bounding box defined by `origin` and `scale`.")

    # fix positions of nodes in first layer
    node_positions = dict()
    node_positions.update(_get_node_positions_within_layer(layers[0], node_spacings[0], layer_positions[0], origin, scale))

    # assign the position of nodes in subsequent layers
    total_layers = len(layers)
    for ii in range(1, total_layers):
        left = layers[ii-1]
        right = layers[ii]
        union = set(left) | set(right)
        edges_between_layers = [(source, target) for (source, target) in edges if (source in union) and (target in union)]
        adjacency_list = _edge_list_to_adjacency_list(edges_between_layers, directed=False)

        # add unconnected nodes
        for node in left:
            adjacency_list.setdefault(node, set())
        for node in right:
            adjacency_list.setdefault(node, set())

        if reduce_edge_crossings:
            if not _is_complete_bipartite(edges, left, right):
                left, right = _reduce_crossings_bipartite(adjacency_list, left, right)

        node_positions.update(_get_node_positions_within_layer(right, node_spacings[ii], layer_positions[ii], origin, scale))

    return node_positions


def _get_node_positions_within_layer(node_order, node_spacing, layer_position, origin, scale):
    node_positions = dict()
    xx = layer_position
    y = node_spacing * np.arange(len(node_spacing))
    y -= np.mean(y)
    y += origin[1] + scale[1] / 2
    for node, yy in zip(node_order, y):
        node_positions[node] = np.array((xx, yy))
    return node_positions


def get_shell_layout(edges, shells, radii=None, origin=(0, 0), scale=(1, 1), pad_by=0.05, reduce_edge_crossings=True):
    """Shell layout.

    This is a wrapper around `get_multipartite_layout` that arranges nodes in shells around a center instead of in layers.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    shells : list
        List of node subsets, one for each shell of the graph.
    radii : list, default None
        List of radii, one for each shell.
        If None, radii are chosen such that the shells are evenly spaced within the bounding box defined by origin and scale.
    origin : tuple
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    reduce_edge_crossings : bool, default True
        If True, attempts to reduce edge crossings via the algorithm outlined in [Eades1994]_.

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Eades1994] Eades & Wormald (1994) Edge crossings in drawings of bipartite graphs.

    """

    if radii is None:
        if len(shells[0]) == 1:
            # Innermost shell consists of a single node, and hence should have no size.
            radii = np.linspace(0, (1 - pad_by) * np.min(scale) / 2, len(shells))
        else:
            # Innermost shell consists of multiple nodes and should hence have a non-zero radius.
            radii = np.linspace(0, (1 - pad_by) * np.min(scale) / 2, len(shells) + 1)[1:]

    relative_radii = np.array(radii) / np.max(radii)
    multipartite_positions = get_multipartite_layout(
        edges, shells, layer_positions=relative_radii,
        reduce_edge_crossings=reduce_edge_crossings,
        uniform_node_spacing=False,
        pad_by=0,
    )

    node_to_shell = {node : shell for shell in shells for node in shell}

    max_radius = np.max(radii)
    center = np.array([origin[0] + 0.5 * scale[0], origin[1] + 0.5 * scale[1]])

    node_positions = dict()
    for node, (x, y) in multipartite_positions.items():
        shell = node_to_shell[node]
        max_angle = 2 * np.pi * len(shell) / (len(shell) + 1)
        angle = max_angle * y
        radius = max_radius * x
        node_positions[node] = _convert_polar_to_cartesian_coordinates(radius, angle) + center

    return node_positions


@_rectangle_pack_multiple_components
def get_community_layout(edges, node_to_community, origin=(0, 0), scale=(1, 1), pad_by=0.05):
    """Community node layout for modular graphs.

    This layout is based on ideas presented in [Traud2009], and involves the following steps:

      1. Position the communities with respect to each other: create a new, weighted graph, where each node corresponds to a community, and the weights correspond to the number of edges between communities, and compute a layout for this community graph.
      2. Position the nodes within each community: for each community, create a subgraph. Find a layout for the subgraph.
      3. Combine positions from step 1 & 2 such that node positions are dominated by the corresponding community position and refined by their position within the community subgraph.
      4. Rotate communities to reduce the length of inter-community edges.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    node_to_community : dict
        The network partition, which maps each node ID to a community ID.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code::

            xmin = origin[0] + pad_by * scale[0]
            ymin = origin[1] + pad_by * scale[1]
            xmax = origin[0] + scale[0] - pad_by * scale[0]
            ymax = origin[1] + scale[1] - pad_by * scale[1]

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    References
    ----------
    .. [Traud2009] Traud et al. (2009) Visualization of communities in networks. https://doi.org/10.1063/1.3194108

    """

    # assert that there multiple communities in the graph; otherwise abort
    nodes = _get_unique_nodes(edges)
    communities = set([node_to_community[node] for node in nodes])
    if len(communities) < 2:
        warnings.warn("Graph contains a single community. Unable to compute a community layout. Computing spring layout instead.")
        return get_fruchterman_reingold_layout(edges, origin=origin, scale=scale, pad_by=pad_by)

    # assert that node_to_community is non-redundant,
    # i.e. only contains nodes that are also present in edges
    node_to_community = {node : node_to_community[node] for node in nodes}

    community_size = _get_community_sizes(node_to_community, scale)
    community_centroids = _get_community_positions(edges, node_to_community, community_size, origin, scale, pad_by)
    relative_node_positions = _get_within_community_positions(edges, node_to_community)
    node_positions = _combine_positions(node_to_community, community_centroids, community_size, relative_node_positions)
    node_positions = _rotate_communities(edges, node_to_community, community_centroids, node_positions)

    return node_positions


def _get_community_sizes(node_to_community, scale):
    """Compute the area of the canvas reserved for each community."""
    total_nodes = len(node_to_community)
    max_radius = np.linalg.norm(scale) / 2
    scalar = max_radius / total_nodes # this is the worst case scenario, where all comunities are lined up like beads on a string; may warrant revisiting
    community_to_nodes = _invert_dict(node_to_community)
    community_size = {community : len(nodes) * scalar for community, nodes in community_to_nodes.items()}
    return community_size


def _get_community_positions(edges, node_to_community, community_size, origin, scale, pad_by):
    """Compute a centroid position for each community."""
    # create a weighted graph, in which each node corresponds to a community,
    # and each edge weight to the number of edges between communities
    between_community_edges = _find_between_community_edges(edges, node_to_community)

    # find layout for communities
    return get_fruchterman_reingold_layout(
        list(between_community_edges.keys()), edge_weight=between_community_edges,
        node_size=community_size, origin=origin, scale=scale, pad_by=pad_by,
    )


def _find_between_community_edges(edges, node_to_community):
    """Convert the graph into a weighted network of communities."""
    between_community_edges = dict()

    for (ni, nj) in edges:
        ci = node_to_community[ni]
        cj = node_to_community[nj]

        if ci != cj:
            if (ci, cj) in between_community_edges:
                between_community_edges[(ci, cj)] += 1
            elif (cj, ci) in between_community_edges:
                # only compute the undirected graph
                between_community_edges[(cj, ci)] += 1
            else:
                between_community_edges[(ci, cj)] = 1

    return between_community_edges


def _get_within_community_positions(edges, node_to_community):
    """Positions nodes within communities."""
    community_to_nodes = _invert_dict(node_to_community)
    node_positions = dict()
    for community, nodes in community_to_nodes.items():
        if len(nodes) > 1:
            subgraph = _get_subgraph(edges, list(nodes))
            if subgraph:
                subgraph_node_positions = get_fruchterman_reingold_layout(
                    subgraph, nodes=nodes, origin=np.array([-1, -1]), scale=np.array([2, 2]))
                node_positions.update(subgraph_node_positions)
            else:
                warnings.warn(f"There are no connections within community {community}. The placement of of nodes within this community is arbitrary.")
                node_positions.update({node : np.random.rand(2) * 2 + np.array([-1, -1]) for node in nodes})
        elif len(nodes) == 1:
            node_positions.update({nodes.pop() : np.array([0., 0.])})
    return node_positions


def _combine_positions(node_to_community, community_centroids, community_size, relative_node_positions):
    node_positions = dict()
    for node, community in node_to_community.items():
        xy = community_centroids[community]
        delta = relative_node_positions[node] * community_size[community]
        node_positions[node] = xy + delta
    return node_positions


def _rotate_communities(edges, node_to_community, community_centroids, node_positions, step_size=0.1, max_iterations=200):

    between_community_edges = [(source, target) for (source, target) in edges \
                               if node_to_community[source] != node_to_community[target]]

    for _ in range(max_iterations):

        # compute torques
        community_torque = {community : 0 for community in set(list(node_to_community.values()))}
        for (source, target) in between_community_edges:
            # source
            community = node_to_community[source]
            r = community_centroids[community] - node_positions[source]
            delta = node_positions[target] - node_positions[source]
            F = delta * np.linalg.norm(delta) # direction * distance**2
            community_torque[community] += np.cross(r, F)

            # target
            community = node_to_community[target]
            r = community_centroids[community] - node_positions[target]
            delta = node_positions[source] - node_positions[target]
            F = delta * np.linalg.norm(delta)
            community_torque[community] += np.cross(r, F)

        # update node positions
        for node, community in node_to_community.items():
            node_positions[node] = _rotate(step_size * -community_torque[community],
                                           node_positions[node],
                                           community_centroids[community])

    return node_positions


def _rotate(angle, points, origin=(0, 0)):
    # https://stackoverflow.com/a/58781388/2912349
    R = np.array([[np.cos(angle), -np.sin(angle)],
                  [np.sin(angle),  np.cos(angle)]])
    origin = np.atleast_2d(origin)
    points = np.atleast_2d(points)
    return np.squeeze((R @ (points.T-origin.T) + origin.T).T)


@_rectangle_pack_multiple_components
def get_geometric_layout(edges, edge_length, node_size=0., tol=1e-3, origin=(0, 0), scale=(1, 1), pad_by=0.05):
    """Node layout for defined edge lengths but unknown node positions.

    Node positions are determined through non-linear optimisation: the
    total distance between nodes is maximised subject to the constraint
    imposed by the edge lengths, which are used as upper bounds.
    If provided, node sizes are used to set lower bounds to minimise collisions.

    ..note:: This implementation is slow.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    edge_lengths : dict
        Mapping of edges to their lengths.
    node_size : scalar or dict, default 0.
        Size (radius) of nodes.
        Providing the correct node size minimises the overlap of nodes in the graph,
        which can otherwise occur if there are many nodes, or if the nodes differ considerably in size.
    tolerance : float, default 1e-3
        The tolerance of the cost function. Small values increase the accuracy, large values improve the computation time.
    origin : tuple, default (0, 0)
        The (float x, float y) coordinates corresponding to the lower left corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1, 1)
        The (float x, float y) dimensions representing the width and height of the bounding box specifying the extent of the canvas.
    pad_by : float, default 0.05
        Padding around node positions to reduce clipping of the node artists with the frame,
        and to create space for routing curved edges including self-loops around nodes.
        This results in the following bounding box:

        .. code-block::

           xmin = origin[0] + pad_by * scale[0]
           ymin = origin[1] + pad_by * scale[1]
           xmax = origin[0] + scale[0] - pad_by * scale[0]
           ymax = origin[1] + scale[1] - pad_by * scale[1]

    Returns
    -------
    node_positions : dict
        Dictionary mapping each node ID to (float x, float y) tuple, the node position.

    """

    # TODO: assert triangle inequality is not violated.
    # HOLD: probably not necessary, as minimisation can still proceed when triangle inequality is violated.

    # assert that the edges fit within the canvas dimensions
    width, height = scale
    max_length = np.sqrt(width**2 + height**2)
    too_long = dict()
    for edge, length in edge_length.items():
        if length > max_length:
            too_long[edge] = length
    if too_long:
        msg = f"The following edges exceed the dimensions of the canvas (`scale={scale}`):"
        for edge, length in too_long.items():
            msg += f"\n\t{edge} : {length}"
        msg += "\nEither increase the `scale` parameter, or decrease the edge lengths."
        raise ValueError(msg)

    # ensure that graph is bi-directional
    edges = edges + [(target, source) for (source, target) in edges] # forces copy
    edges = list(set(edges))

    # upper bound: pairwise distance matrix with unknown distances set to the maximum possible distance given the canvas dimensions

    lengths = []
    for (source, target) in edges:
        if (source, target) in edge_length:
            lengths.append(edge_length[(source, target)])
        else:
            lengths.append(edge_length[(target, source)])

    sources, targets = zip(*edges)
    nodes = sources + targets
    unique_nodes = set(nodes)
    indices = range(len(unique_nodes))
    node_to_idx = dict(zip(unique_nodes, indices))
    source_indices = [node_to_idx[source] for source in sources]
    target_indices = [node_to_idx[target] for target in targets]

    total_nodes = len(unique_nodes)
    distance_matrix = np.full((total_nodes, total_nodes), max_length)
    distance_matrix[source_indices, target_indices] = lengths
    distance_matrix[np.diag_indices(total_nodes)] = 0
    upper_bounds = squareform(distance_matrix)

    # lower bound: sum of node sizes
    if isinstance(node_size, (int, float)):
        sizes = node_size * np.ones((total_nodes))
    elif isinstance(node_size, dict):
        sizes = np.array([node_size[node] if node in node_size else 0. for node in unique_nodes])

    sum_of_node_sizes = sizes[np.newaxis, :] + sizes[:, np.newaxis]
    sum_of_node_sizes -= np.diag(np.diag(sum_of_node_sizes)) # squareform requires zeros on diagonal
    lower_bounds = squareform(sum_of_node_sizes)
    invalid = lower_bounds > upper_bounds
    lower_bounds[invalid] = upper_bounds[invalid] - 1e-8

    # For an extended discussion of this cost function and alternatives see:
    # https://stackoverflow.com/q/75137677/2912349
    def cost_function(positions):
        return 1 / np.sum(np.log(pdist(positions.reshape((-1, 2))) + 1))

    def constraint_function(positions):
        positions = np.reshape(positions, (-1, 2))
        return pdist(positions)

    initial_positions = _initialise_geometric_node_layout(edges, edge_length)
    nonlinear_constraint = NonlinearConstraint(constraint_function, lb=lower_bounds, ub=upper_bounds, jac='2-point')
    result = minimize(
        cost_function,
        initial_positions.flatten(),
        method='SLSQP',
        jac='2-point',
        constraints=[nonlinear_constraint],
        options=dict(ftol=tol),
    )

    if not result.success:
        print("Warning: could not compute valid node positions for the given edge lengths.")
        print(f"scipy.optimize.minimize: {result.message}.")

    node_positions_as_array = result.x.reshape((-1, 2))
    node_positions_as_array = _fit_to_frame(node_positions_as_array, np.array(origin), np.array(scale), pad_by)
    node_positions = dict(zip(unique_nodes, node_positions_as_array))
    return node_positions


def _initialise_geometric_node_layout(edges, edge_length):
    """Initialises the node positions using the FR algorithm with weights.
    Shorter edges are given a larger weight such that the nodes experience a strong attractive force."""

    edge_weight = dict()
    for edge, length in edge_length.items():
        edge_weight[edge] = 1 / length
    node_positions = get_fruchterman_reingold_layout(edges, edge_weight=edge_weight)
    return np.array(list(node_positions.values()))


def _fit_to_frame(positions, origin, scale, pad_by):
    """Rotate, rescale and shift a set of positions such that they fit
    inside a frame while preserving the relative distances between
    them."""

    # find major axis
    delta = positions[np.newaxis, :] - positions[:, np.newaxis]
    distances = np.sum(delta**2, axis=-1)
    ii, jj = np.where(np.triu(distances)==np.max(distances))

    # use the first if there are several solutions
    ii = ii[0]
    jj = jj[0]

    # pivot around half-way point
    pivot = positions[ii] + 0.5 * delta[ii, jj]
    angle = _get_angle(*delta[ii, jj])

    if scale[0] < scale[1]: # portrait
        rotated_positions = _rotate((np.pi/2 - angle) % np.pi, positions, pivot)
    else: # landscape
        rotated_positions = _rotate(-angle % np.pi, positions, pivot)

    # shift to (0, 0)
    shifted_positions = rotated_positions - np.min(rotated_positions, axis=0)[np.newaxis, :]

    # rescale & center
    dx, dy = np.ptp(rotated_positions, axis=0)
    if dx/scale[0] < dy/scale[1]:
        rescaled_positions = shifted_positions * (1 - 2 * pad_by) * scale[1] / dy
        rescaled_positions[:, 0] += (scale[0] - np.ptp(rescaled_positions[:, 0])) / 2
        rescaled_positions[:, 1] += pad_by * scale[1]
    else:
        rescaled_positions = shifted_positions * (1 - 2 * pad_by) * scale[0] / dx
        rescaled_positions[:, 0] += pad_by * scale[0]
        rescaled_positions[:, 1] += (scale[1] - np.ptp(rescaled_positions[:, 1])) / 2

    # shift to origin
    reshifted_positions = rescaled_positions + np.array(origin)[np.newaxis, :]

    return reshifted_positions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implements the InteractiveGraph, MutableGraph, and EditableGraph classes,
as well as the following helper classes:

  - ClickableArtists
  - SelectableArtists
  - DraggableArtists
  - DraggableGraph
  - DraggableGraphWithGridMode
  - EmphasizeOnHover
  - EmphasizeOnHoverGraph
  - AnnotateOnClick
  - AnnotateOnClickGraph
  - TableOnClick
  - TableOnClickGraph

"""

import warnings
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backend_bases import key_press_handler

from ._graph_classes import BASE_SCALE, Graph
from ._artists import (
    NodeArtist,
    CircularNodeArtist,
    RegularPolygonNodeArtist,
    EdgeArtist,
)
from ._utils import (
    _get_angle,
    _get_interior_angle_between,
    _get_orthogonal_unit_vector,
    _get_point_along_spline,
    _get_tangent_at_point,
)
from ._parser import is_order_zero, is_empty, parse_graph


DEFAULT_EDGE = (0, 1)
DEFAULT_EDGE_KEY = 0


class ClickableArtists(object):
    """Implements selection of matplotlib artists via the mouse left click (+/- ctrl or command key).

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/47312637/2912349

    """
    def __init__(self, artists):

        try:
            self.fig, = set(list(artist.figure for artist in artists))
        except ValueError:
            raise Exception("All artists have to be on the same figure!")

        try:
            self.ax, = set(list(artist.axes for artist in artists))
        except ValueError:
            raise Exception("All artists have to be on the same axis!")

        # self.fig.canvas.mpl_connect('button_press_event', self._on_press)
        self.fig.canvas.mpl_connect('button_release_event', self._on_release)

        self._clickable_artists = list(artists)
        self._selected_artists = []
        self._base_linewidth = dict([(artist, artist._lw_data) for artist in artists])
        self._base_edgecolor = dict([(artist, artist.get_edgecolor()) for artist in artists])

        if plt.get_backend() == 'MacOSX':
            msg  = "You appear to be using the MacOSX backend."
            msg += "\nModifier key presses are bugged on this backend. See https://github.com/matplotlib/matplotlib/issues/20486"
            msg += "\nConsider using a different backend, e.g. TkAgg (import matplotlib; matplotlib.use('TkAgg'))."
            msg += "\nNote that you must set the backend before importing any package depending on matplotlib (includes pyplot, networkx, netgraph)."
            warnings.warn(msg)

    # def _on_press(self, event):
    def _on_release(self, event):
        if event.inaxes == self.ax:
            for artist in self._clickable_artists:
                if artist.contains(event)[0]:
                    if event.key in ('control', 'super+??', 'ctrl+??'):
                        self._toggle_select_artist(artist)
                    else:
                        self._deselect_all_other_artists(artist)
                        self._toggle_select_artist(artist)
                        # NOTE: if two artists are overlapping, only the first one encountered is selected!
                    break
            else:
                if not event.key in ('control', 'super+??', 'ctrl+??'):
                    self._deselect_all_artists()
        else:
            print("Warning: clicked outside axis limits!")


    def _toggle_select_artist(self, artist):
        if artist in self._selected_artists:
            self._deselect_artist(artist)
        else:
            self._select_artist(artist)


    def _select_artist(self, artist):
        if not (artist in self._selected_artists):
            linewidth = artist._lw_data
            artist.set_linewidth(max(1.5 * linewidth, 0.003))
            artist.set_edgecolor('black')
            self._selected_artists.append(artist)
            self.fig.canvas.draw_idle()


    def _deselect_artist(self, artist):
        if artist in self._selected_artists: # should always be true?
            artist.set_linewidth(self._base_linewidth[artist])
            artist.set_edgecolor(self._base_edgecolor[artist])
            self._selected_artists.remove(artist)
            self.fig.canvas.draw_idle()


    def _deselect_all_artists(self):
        for artist in self._selected_artists[:]: # we make a copy of the list with [:], as we are modifying the list being iterated over
            self._deselect_artist(artist)


    def _deselect_all_other_artists(self, artist_to_keep):
        for artist in self._selected_artists[:]:
            if artist != artist_to_keep:
                self._deselect_artist(artist)


class SelectableArtists(ClickableArtists):
    """Augments :py:class:`ClickableArtists` with a rectangle selector.

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/47312637/2912349

    """
    def __init__(self, artists):
        super().__init__(artists)

        self.fig.canvas.mpl_connect('button_press_event', self._on_press)
        # self.fig.canvas.mpl_connect('button_release_event', self._on_release)
        self.fig.canvas.mpl_connect('motion_notify_event',  self._on_motion)

        self._selectable_artists = list(artists)
        self._currently_selecting = False

        self._rect = plt.Rectangle((0, 0), 1, 1, linestyle="--", edgecolor="crimson", fill=False)
        self.ax.add_patch(self._rect)
        self._rect.set_visible(False)

        self._x0 = 0
        self._y0 = 0
        self._x1 = 0
        self._y1 = 0


    def _on_press(self, event):
        # super()._on_press(event)

        if event.inaxes == self.ax:
            # reset rectangle
            self._x0 = event.xdata
            self._y0 = event.ydata
            self._x1 = event.xdata
            self._y1 = event.ydata

            for artist in self._clickable_artists:
                if artist.contains(event)[0]:
                    break
            else:
                self._currently_selecting = True


    def _on_release(self, event):
        super()._on_release(event)

        if self._currently_selecting:
            # select artists inside window
            for artist in self._selectable_artists:
                if isinstance(artist, NodeArtist):
                    if self._is_inside_rect(*artist.xy):
                        if event.key in ('control', 'super+??', 'ctrl+??'): # if/else probably superfluouos
                            self._toggle_select_artist(artist)              # as no artists will be selected
                        else:                                               # if control is not held previously
                            self._select_artist(artist)                     #
                elif isinstance(artist, EdgeArtist):
                    if np.all([self._is_inside_rect(x, y) for x, y in artist.midline]):
                        if event.key in ('control', 'super+??', 'ctrl+??'): # if/else probably superfluouos
                            self._toggle_select_artist(artist)              # as no artists will be selected
                        else:                                               # if control is not held previously
                            self._select_artist(artist)                     #

            # stop window selection and draw new state
            self._currently_selecting = False
            self._rect.set_visible(False)
            self.fig.canvas.draw_idle()


    def _on_motion(self, event):
        if event.inaxes == self.ax:
            if self._currently_selecting:
                self._x1 = event.xdata
                self._y1 = event.ydata
                # add rectangle for selection here
                self._selector_on()


    def _is_inside_rect(self, x, y):
        xlim = np.sort([self._x0, self._x1])
        ylim = np.sort([self._y0, self._y1])
        if (xlim[0]<=x) and (x<xlim[1]) and (ylim[0]<=y) and (y<ylim[1]):
            return True
        else:
            return False


    def _selector_on(self):
        self._rect.set_visible(True)
        xlim = np.sort([self._x0, self._x1])
        ylim = np.sort([self._y0, self._y1])
        self._rect.set_xy((xlim[0], ylim[0]))
        self._rect.set_width(xlim[1] - xlim[0])
        self._rect.set_height(ylim[1] - ylim[0])
        self.fig.canvas.draw_idle()


class DraggableArtists(SelectableArtists):
    """Augments :py:class:`SelectableArtists` to support dragging of artists by holding the left mouse button.

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/47312637/2912349

    """

    def __init__(self, artists):
        super().__init__(artists)

        self._draggable_artists = list(artists)
        self._currently_clicking_on_artist = None
        self._currently_dragging = False
        self._offset = dict()


    def _on_press(self, event):
        super()._on_press(event)

        if event.inaxes == self.ax:
            for artist in self._draggable_artists:
                if artist.contains(event)[0]:
                    self._currently_clicking_on_artist = artist
                    break
        else:
            print("Warning: clicked outside axis limits!")


    def _on_motion(self, event):
        super()._on_motion(event)

        if event.inaxes == self.ax:
            if self._currently_clicking_on_artist:
                if self._currently_clicking_on_artist not in self._selected_artists:
                    if event.key not in ('control', 'super+??', 'ctrl+??'):
                        self._deselect_all_artists()
                    self._select_artist(self._currently_clicking_on_artist)
                self._offset = {artist : artist.xy - np.array([event.xdata, event.ydata]) for artist in self._selected_artists if artist in self._draggable_artists}
                self._currently_clicking_on_artist = None
                self._currently_dragging = True

            if self._currently_dragging:
                self._move(event)


    def _on_release(self, event):
        if self._currently_dragging:
            self._currently_dragging = False
        else:
            self._currently_clicking_on_artist = None
            super()._on_release(event)


    def _move(self, event):
        cursor_position = np.array([event.xdata, event.ydata])
        for artist in self._selected_artists:
            artist.xy = cursor_position + self._offset[artist]
        self.fig.canvas.draw_idle()


class DraggableGraph(Graph, DraggableArtists):
    """
    Augments :py:class:`Graph` to support selection and dragging of node artists with the mouse.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    *args, **kwargs
        Parameters passed through to :py:class:`Graph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`DraggableArtists`, :py:class:`InteractiveGraph`

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_dragging_clicking_and_selecting()


    def _setup_dragging_clicking_and_selecting(self):
        DraggableArtists.__init__(self, self.node_artists.values())
        self._draggable_artist_to_node = dict(zip(self.node_artists.values(), self.node_artists.keys()))
        self._clickable_artists.extend(list(self.edge_artists.values()))
        self._selectable_artists.extend(list(self.edge_artists.values()))
        self._base_linewidth.update(dict([(artist, artist._lw_data) for artist in self.edge_artists.values()]))
        self._base_edgecolor.update(dict([(artist, artist.get_edgecolor()) for artist in self.edge_artists.values()]))
        # # trigger resize of labels when canvas size changes
        # self.fig.canvas.mpl_connect('resize_event', self._on_resize)


    def _move(self, event):

        cursor_position = np.array([event.xdata, event.ydata])

        nodes = self._get_stale_nodes()
        self._update_node_positions(nodes, cursor_position)
        self._update_node_artists(nodes)

        if hasattr(self, 'node_label_artists'):
            self._update_node_label_positions()

        edge_paths = self.edge_layout.get(nodes, approximate=True)
        self._update_edge_artists(edge_paths)
        if hasattr(self, 'edge_label_artists'):
            self._update_edge_label_positions(edge_paths.keys())

        self.fig.canvas.draw_idle()


    def _get_stale_nodes(self):
        return [self._draggable_artist_to_node[artist] for artist in self._selected_artists if artist in self._draggable_artists]


    def _update_node_positions(self, nodes, cursor_position):
        for node in nodes:
            self.node_positions[node] = cursor_position + self._offset[self.node_artists[node]]


    def _on_release(self, event):
        if self._currently_dragging:
            edge_paths = self.edge_layout.get()
            self._update_edge_artists(edge_paths)
            if hasattr(self, 'edge_label_artists'): # move edge labels
                self._update_edge_label_positions(edge_paths.keys())

        super()._on_release(event)


    # def _on_resize(self, event):
    #     if hasattr(self, 'node_labels'):
    #         self.draw_node_labels(self.node_labels)
    #         # print("As node label font size was not explicitly set, automatically adjusted node label font size to {:.2f}.".format(self.node_label_font_size))


class DraggableGraphWithGridMode(DraggableGraph):
    """
    Adds a grid-mode to :py:class:`DraggableGraph`, in which node positions are fixed to a grid.
    To activate, press the letter 'g'.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    *args, **kwargs
        Parameters passed through to :py:class:`Graph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`DraggableGraph`, :py:class:`InteractiveGraph`

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_grid_mode()


    def _setup_grid_mode(self):
        self.grid = False
        self.grid_dx = 0.05 * self.scale[0]
        self.grid_dy = 0.05 * self.scale[1]
        self._grid_lines = []
        self.fig.canvas.mpl_connect('key_press_event', self._on_key_toggle)


    def _on_key_toggle(self, event):
        if event.key == 'g':
            if self.grid is False:
                self.grid = True
                self._draw_grid()
            else:
                self.grid = False
                self._remove_grid()
        self.fig.canvas.draw_idle()


    def _draw_grid(self):
        eps = 1e-13
        for x in np.arange(self.origin[0], self.origin[0] + self.scale[0] + eps, self.grid_dx):
            line = self.ax.axvline(x, color='k', alpha=0.1, linestyle='--')
            self._grid_lines.append(line)

        for y in np.arange(self.origin[1], self.origin[1] + self.scale[1] + eps, self.grid_dy):
            line = self.ax.axhline(y, color='k', alpha=0.1, linestyle='--')
            self._grid_lines.append(line)


    def _remove_grid(self):
        for line in self._grid_lines:
            line.remove()
        self._grid_lines = []


    def _on_release(self, event):
        if self._currently_dragging and self.grid:
            nodes = self._get_stale_nodes()
            for node in nodes:
                self.node_positions[node] = self._get_nearest_grid_coordinate(*self.node_positions[node])
            self._update_node_artists(nodes)
            if hasattr(self, 'node_label_artists'):
                self._update_node_label_positions()

            edge_paths = self.edge_layout.get()
            self._update_edge_artists(edge_paths)
            if hasattr(self, 'edge_label_artists'): # move edge labels
                self._update_edge_label_positions(edges)

        super()._on_release(event)


    def _get_nearest_grid_coordinate(self, x, y):
        x = np.round((x - self.origin[0]) / self.grid_dx) * self.grid_dx + self.origin[0]
        y = np.round((y - self.origin[1]) / self.grid_dy) * self.grid_dy + self.origin[1]
        return x, y


class EmphasizeOnHover(object):
    """Emphasize matplotlib artists when hovering over them by desaturating all other artists."""

    def __init__(self, artists):

        self.emphasizeable_artists = artists
        self._base_alpha = {artist : artist.get_alpha() for artist in self.emphasizeable_artists}
        self.deemphasized_artists = []

        try:
            self.fig, = set(list(artist.figure for artist in artists))
        except ValueError:
            raise Exception("All artists have to be on the same figure!")

        try:
            self.ax, = set(list(artist.axes for artist in artists))
        except ValueError:
            raise Exception("All artists have to be on the same axis!")

        self.fig.canvas.mpl_connect("motion_notify_event", self._on_motion)


    def _on_motion(self, event):
        if event.inaxes == self.ax:
            artist = self._is_on_artist(event)
            if artist:
                self._emphasize(artist)
            elif self.deemphasized_artists:
                self._reset_emphasis()


    def _is_on_artist(self, event):
        for artist in self.emphasizeable_artists:
            if artist.contains(event)[0]: # returns two arguments for some reason
                return artist
        return None


    def _emphasize(self, artist_to_emphasize):
        for artist in self.emphasizeable_artists:
            if artist is not artist_to_emphasize:
                artist.set_alpha(self._base_alpha[artist]/5)
                self.deemphasized_artists.append(artist)
        self.fig.canvas.draw_idle()


    def _reset_emphasis(self):
        for artist in self.deemphasized_artists:
            try:
                artist.set_alpha(self._base_alpha[artist])
            except KeyError:
                # This is a workaround for a bug in MutableGraph that occurs when:
                # 1) The user selects an artist.
                # 2) The user moves the mouse and hovers over a different artist that results in de-emphasizing the selected artist.
                # 3) The user deletes the selected artist.
                # In theory, self.deemphasized_artists should be updated on node/edge removal.
                # In practice, it still contains the artist when this function is called.
                pass
        self.deemphasized_artists.clear()
        self.fig.canvas.draw_idle()


class EmphasizeOnHoverGraph(Graph, EmphasizeOnHover):
    """Combines :py:class:`EmphasizeOnHover` with the :py:class:`Graph` class
    such that nodes are emphasized when hovering over them with the mouse.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    mouseover_highlight_mapping : dict or None, default None
        Determines which nodes and/or edges are highlighted when hovering over any given node or edge.
        The keys of the dictionary are node and/or edge IDs, while the values are iterables of node and/or edge IDs.
        If the parameter is None, a default dictionary is constructed, which maps

        - edges to themselves as well as their source and target nodes, and
        - nodes to themselves as well as their immediate neighbours and any edges between them.

    *args, **kwargs
        Parameters passed through to :py:class:`Graph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`EmphasizeOnHover`, :py:class:`InteractiveGraph`

    """

    def __init__(self, graph, mouseover_highlight_mapping=None, *args, **kwargs):
        Graph.__init__(self, graph, *args, **kwargs)

        self._setup_emphasis()
        if mouseover_highlight_mapping is not None: # this includes empty mappings!
            self._check_mouseover_highlight_mapping(mouseover_highlight_mapping)
            self.mouseover_highlight_mapping = mouseover_highlight_mapping


    def _setup_emphasis(self):
        self.artist_to_key = self._map_artist_to_key()
        EmphasizeOnHover.__init__(self, list(self.artist_to_key.keys()))
        self.mouseover_highlight_mapping = self._get_default_mouseover_highlight_mapping()


    def _map_artist_to_key(self):
        artists = list(self.node_artists.values()) + list(self.edge_artists.values())
        keys = list(self.node_artists.keys()) + list(self.edge_artists.keys())
        return dict(zip(artists, keys))


    def _get_default_mouseover_highlight_mapping(self):
        """Construct default mapping:

        - Nodes map to themselves, their neighbours, and any edges between them.
        - Edges map to themselves, and their source and target node.

        """

        mapping = dict()

        for node in self.nodes:
            mapping[node] = [node]

        for edge in self.edges:
            source, target = edge[:2]
            mapping[source].append(edge)
            mapping[source].append(target)
            mapping[target].append(edge)
            mapping[target].append(source)
            mapping[edge] = [edge, source, target]

        return mapping


    def _check_mouseover_highlight_mapping(self, mapping):

        if not isinstance(mapping, dict):
            raise TypeError(f"Parameter `mouseover_highlight_mapping` is a dictionary, not {type(mapping)}.")

        invalid_keys = []
        for key in mapping:
            if key in self.node_artists:
                pass
            elif key in self.edge_artists:
                pass
            else:
                invalid_keys.append(key)
        if invalid_keys:
            msg = "Parameter `mouseover_highlight_mapping` contains invalid keys:"
            for key in invalid_keys:
                msg += f"\n\t- {key}"
            raise ValueError(msg)

        invalid_values = []
        for values in mapping.values():
            for value in values:
                if value in self.node_artists:
                    pass
                elif value in self.edge_artists:
                    pass
                else:
                    invalid_values.append(value)
        if invalid_values:
            msg = "Parameter `mouseover_highlight_mapping` contains invalid values:"
            for value in set(invalid_values):
                msg += f"\n\t- {value}"
            raise ValueError(msg)


    def _emphasize(self, selected_artist):
        key = self.artist_to_key[selected_artist]
        if key in self.mouseover_highlight_mapping:
            emphasized_artists = []
            for value in self.mouseover_highlight_mapping[key]:
                if value in self.node_artists:
                    emphasized_artists.append(self.node_artists[value])
                elif value in self.edge_artists:
                    emphasized_artists.append(self.edge_artists[value])

            for artist in self.emphasizeable_artists:
                if artist not in emphasized_artists:
                    artist.set_alpha(self._base_alpha[artist]/5)
                    self.deemphasized_artists.append(artist)
            self.fig.canvas.draw_idle()


class AnnotateOnClick(object):
    """Show or hide annotations when clicking on matplotlib artists."""

    def __init__(self, artist_to_annotation, annotation_fontdict=None):
        self.artist_to_annotation = artist_to_annotation
        self.annotated_artists = set()
        self.artist_to_text_object = dict()
        self.annotation_fontdict = dict(
            backgroundcolor = 'white',
            zorder          = np.inf,
            clip_on         = False
        )
        if annotation_fontdict:
            self.annotation_fontdict.update(annotation_fontdict)

        if not hasattr(self, "fig"):
            try:
                self.fig, = set(list(artist.figure for artist in artist_to_annotation))
            except ValueError:
                raise Exception("All artists have to be on the same figure!")

        if not hasattr(self, "ax"):
            try:
                self.ax, = set(list(artist.axes for artist in artist_to_annotation))
            except ValueError:
                raise Exception("All artists have to be on the same axis!")

        self.fig.canvas.mpl_connect("button_release_event", self._on_release)


    def _on_release(self, event):
        if event.inaxes == self.ax:

            # clicked on already annotated artist
            for artist in self.annotated_artists:
                if artist.contains(event)[0]:
                    self._remove_annotation(artist)
                    self.fig.canvas.draw()
                    return

            # clicked on un-annotated artist
            for artist in self.artist_to_annotation:
                if artist.contains(event)[0]:
                    placement = self._get_annotation_placement(artist)
                    self._add_annotation(artist, *placement)
                    self.fig.canvas.draw()
                    return

            # # clicked outside of any artist
            # for artist in list(self.annotated_artists): # list to force copy
            #     self._remove_annotation(artist)
            # self.fig.canvas.draw()


    def _get_annotation_placement(self, artist):
        vector = self._get_vector_pointing_outwards(artist.xy)
        x, y = artist.xy + 2 * artist.size * vector
        horizontalalignment, verticalalignment = self._get_text_alignment(vector)
        return x, y, horizontalalignment, verticalalignment


    def _get_centroid(self):
        return np.mean([artist.xy for artist in self.artist_to_annotation], axis=0)


    def _get_vector_pointing_outwards(self, xy):
        centroid = self._get_centroid()
        delta = xy - centroid
        distance = np.linalg.norm(delta)
        unit_vector = delta / distance
        return unit_vector


    def _get_text_alignment(self, vector):
        dx, dy = vector
        angle = _get_angle(dx, dy, radians=False) % 360

        if (45 <= angle < 135):
            horizontalalignment = 'center'
            verticalalignment = 'bottom'
        elif (135 <= angle < 225):
            horizontalalignment = 'right'
            verticalalignment = 'center'
        elif (225 <= angle < 315):
            horizontalalignment = 'center'
            verticalalignment = 'top'
        else:
            horizontalalignment = 'left'
            verticalalignment = 'center'

        return horizontalalignment, verticalalignment


    def _add_annotation(self, artist, x, y, horizontalalignment, verticalalignment):
        params = self.annotation_fontdict.copy()
        params.setdefault('horizontalalignment', horizontalalignment)
        params.setdefault('verticalalignment', verticalalignment)
        if isinstance(self.artist_to_annotation[artist], str):
            self.artist_to_text_object[artist] = self.ax.text(
                x, y, self.artist_to_annotation[artist], **params)
        elif isinstance(self.artist_to_annotation[artist], dict):
            params.update(self.artist_to_annotation[artist].copy())
            self.artist_to_text_object[artist] = self.ax.text(
                x, y, **params
            )
        self.annotated_artists.add(artist)


    def _remove_annotation(self, artist):
        text_object = self.artist_to_text_object[artist]
        text_object.remove()
        del self.artist_to_text_object[artist]
        self.annotated_artists.discard(artist)


    def _redraw_annotations(self, event):
        if event.inaxes == self.ax:
            for artist in self.annotated_artists:
                self._remove_annotation(artist)
                placement = self._get_annotation_placement(artist)
                self._add_annotation(artist, *placement)
            self.fig.canvas.draw()


class AnnotateOnClickGraph(Graph, AnnotateOnClick):
    """Combines :py:class:`AnnotateOnClick` with the :py:class:`Graph` class
    such that nodes or edges can have toggleable annotations.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    annotations : dict
        Mapping of nodes or edges to strings or dictionaries, the annotations.
        The visibility of the annotations can be toggled on or off by clicking on the corresponding node or edge.

        .. code-block::

           annotations = {
               0      : 'Normal node',
               1      : {s : 'Less important node', fontsize : 2},
               2      : {s : 'Very important node', fontcolor : 'red'},
               (0, 1) : 'Normal edge',
               (1, 2) : {s : 'Less important edge', fontsize : 2},
               (2, 0) : {s : 'Very important edge', fontcolor : 'red'},
           }

    annotation_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text if only the annotation string is given.
        For a full list of available arguments see the matplotlib documentation.
    *args, **kwargs
        Parameters passed through to :py:class:`Graph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`AnnotateOnClick`, :py:class:`InteractiveGraph`

    """

    def __init__(self, *args, **kwargs):
        Graph.__init__(self, *args, **kwargs)
        self._setup_annotations(*args, **kwargs)


    def _setup_annotations(self, *args, **kwargs):
        if "annotations" in kwargs:
            artist_to_annotation = self._map_artist_to_annotation(kwargs["annotations"])
        else:
            artist_to_annotation = dict()

        if "annotation_fontdict" in kwargs:
            AnnotateOnClick.__init__(self, artist_to_annotation, kwargs["annotation_fontdict"])
        else:
            AnnotateOnClick.__init__(self, artist_to_annotation)


    def _map_artist_to_annotation(self, annotations):
        artist_to_annotation = dict()
        for key, annotation in annotations.items():
            if key in self.nodes:
                artist_to_annotation[self.node_artists[key]] = annotation
            elif key in self.edges:
                artist_to_annotation[self.edge_artists[key]] = annotation
            else:
                raise ValueError(f"There is no node or edge with the ID {key} for the annotation '{annotation}'.")
        return artist_to_annotation


    def _get_centroid(self):
        return Graph._get_centroid(self)


    def _get_annotation_placement(self, artist):
        if isinstance(artist, NodeArtist):
            return self._get_node_annotation_placement(artist)
        elif isinstance(artist, EdgeArtist):
            return self._get_edge_annotation_placement(artist)
        else:
            raise NotImplementedError


    def _get_node_annotation_placement(self, artist):
        return super()._get_annotation_placement(artist)


    def _get_edge_annotation_placement(self, artist):
        midpoint = _get_point_along_spline(artist.midline, 0.5)

        tangent = _get_tangent_at_point(artist.midline, 0.5)
        orthogonal_vector = _get_orthogonal_unit_vector(np.atleast_2d(tangent)).ravel()
        vector_pointing_outwards = self._get_vector_pointing_outwards(midpoint)
        if _get_interior_angle_between(orthogonal_vector, vector_pointing_outwards, radians=False) > 90:
            orthogonal_vector *= -1

        x, y = midpoint + 2 * artist.width * orthogonal_vector
        horizontalalignment, verticalalignment = self._get_text_alignment(orthogonal_vector)
        return x, y, horizontalalignment, verticalalignment


class TableOnClick(object):
    """Show or hide tabular information when clicking on matplotlib artists."""

    def __init__(self, artist_to_table, table_kwargs=None):

        self.artist_to_table = artist_to_table
        self.table = None
        self.table_fontsize = None
        self.table_kwargs = dict(
            # bbox = [1.1, 0.1, 0.5, 0.8],
            # edges = 'horizontal',
        )

        if table_kwargs:
            if 'fontsize' in table_kwargs:
                self.table_fontsize = table_kwargs['fontsize']
            self.table_kwargs.update(table_kwargs)

        if not hasattr(self, "fig"):
            try:
                self.fig, = set(list(artist.figure for artist in artist_to_table))
            except ValueError:
                raise Exception("All artists have to be on the same figure!")

        if not hasattr(self, "ax"):
            try:
                self.ax, = set(list(artist.axes for artist in artist_to_table))
            except ValueError:
                raise Exception("All artists have to be on the same axis!")

        self.fig.canvas.mpl_connect("button_release_event", self._on_release)


    def _on_release(self, event):
        if event.inaxes == self.ax:
            for artist in self.artist_to_table:
                if artist.contains(event)[0]:
                    if self.table:
                        self._remove_table()
                    self._add_table(artist)
                    self.fig.canvas.draw()
                    break
            else:
                if self.table:
                    self._remove_table()
                    self.fig.canvas.draw()


    def _add_table(self, artist):
        df = self.artist_to_table[artist]
        self.table = self.ax.table(
            cellText = df.values.tolist(),
            rowLabels = df.index.values,
            colLabels = df.columns.values,
            **self.table_kwargs,
        )

        if self.table_fontsize:
            self.table.auto_set_font_size(False)
            self.table.set_fontsize(self.table_fontsize)

    def _remove_table(self):
        self.table.remove()
        self.table = None


class TableOnClickGraph(Graph, TableOnClick):
    """Combines :py:class:`TableOnClick` with :py:class:`Graph`
    such that nodes or edges can have toggleable tabular annotations.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    tables : dict node/edge : pandas dataframe
        Mapping of nodes and/or edges to pandas dataframes.
        The visibility of the tables that can toggled on or off by clicking on the corresponding node or edge.
    table_kwargs : dict
        Keyword arguments passed to matplotlib.pyplot.table.
    *args, **kwargs
        Parameters passed through to :py:class:`Graph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`TableOnClick`, :py:class:`InteractiveGraph`

    """

    def __init__(self, *args, **kwargs):
        Graph.__init__(self, *args, **kwargs)
        self._setup_table_annotations(*args, **kwargs)


    def _setup_table_annotations(self, *args, **kwargs):
        if 'tables' in kwargs:
            artist_to_table = self._map_artist_to_table(kwargs['tables'])
        else:
            artist_to_table = dict()

        if 'table_kwargs' in kwargs:
            TableOnClick.__init__(self, artist_to_table, kwargs['table_kwargs'])
        else:
            TableOnClick.__init__(self, artist_to_table)


    def _map_artist_to_table(self, tables):
        artist_to_table = dict()
        for key, table in tables.items():
            if key in self.nodes:
                artist_to_table[self.node_artists[key]] = table
            elif key in self.edges:
                artist_to_table[self.edge_artists[key]] = table
            else:
                raise ValueError(f"There is no node or edge with the ID {key} for the table '{table}'.")
        return artist_to_table


class InteractiveGraph(DraggableGraphWithGridMode, EmphasizeOnHoverGraph, AnnotateOnClickGraph, TableOnClickGraph):
    """Extends the :py:class:`Graph` to support node placement with the mouse,
    emphasis of graph elements when hovering over them, and toggleable annotations.

    - Nodes can be selected and dragged around with the mouse.
    - Nodes and edges are emphasized when hovering over them.
    - Supports additional annotations that can be toggled on and off by clicking on the corresponding node or edge.
    - These annotations can also be tables.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    node_layout : str or dict, default 'spring'
        If a string, the node positions are computed using the indicated method:

        - 'random'       : place nodes in random positions;
        - 'circular'     : place nodes regularly spaced on a circle;
        - 'spring'       : place nodes using a force-directed layout (Fruchterman-Reingold algorithm);
        - 'dot'          : place nodes using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'radial'       : place nodes radially using the Sugiyama algorithm; the graph should be directed and acyclic;
        - 'community'    : place nodes such that nodes belonging to the same community are grouped together;
        - 'bipartite'    : place nodes regularly spaced on two parallel lines;
        - 'multipartite' : place nodes regularly spaced on several parallel lines;
        - 'shell'        : place nodes regularly spaced on concentric circles;
        - 'geometric'    : place nodes according to the length of the edges between them.

        If a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_random_layout`
        - :py:func:`get_circular_layout`
        - :py:func:`get_fruchterman_reingold_layout`
        - :py:func:`get_sugiyama_layout`
        - :py:func:`get_radial_tree_layout`
        - :py:func:`get_community_layout`
        - :py:func:`get_bipartite_layout`
        - :py:func:`get_multipartite_layout`
        - :py:func:`get_shell_layout`
        - :py:func:`get_geometric_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_cmap : matplotlib color map (default 'RdGy')
        Color map used to map edge weights to edge colors. Should be diverging.
        If edge weights are strictly positive, weights are mapped to the
        left hand side of the color map with vmin=0 and vmax=np.max(weights).
        If edge weights are positive and negative, then weights are mapped
        to colors such that a weight of zero corresponds to the center of the
        color map; the boundaries are set to +/- the maximum absolute weight.
        If the graph is unweighted or the edge colors are specified explicitly,
        this parameter is ignored.
    edge_color : matplotlib color specification or dict, default DEFAULT_COLOR
        Edge color. If provided explicitly, overrides `edge_cmap`.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout : str or dict (default 'straight')
        If edge_layout is a string, determine the layout internally:

        - 'straight' : draw edges as straight lines
        - 'curved'   : draw edges as curved splines; the spline control points are optimised to avoid other nodes and edges
        - 'bundled'  : draw edges as edge bundles
        - 'arc'      : draw edges as arcs with a fixed curvature

        If edge_layout is a dict, the keys are edges and the values are edge paths
        in the form iterables of (x, y) tuples, the edge segments.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to edge layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_straight_edge_paths`
        - :py:func:`get_curved_edge_paths`
        - :py:func:`get_bundled_edge_paths`
        - :py:func:`get_arced_edge_paths`

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().
    annotations : dict
        Mapping of nodes or edges to strings or dictionaries, the annotations.
        The visibility of the annotations can be toggled on or off by clicking on the corresponding node or edge.

        .. code-block::

           annotations = {
               0      : 'Normal node',
               1      : {s : 'Less important node', fontsize : 2},
               2      : {s : 'Very important node', fontcolor : 'red'},
               (0, 1) : 'Normal edge',
               (1, 2) : {s : 'Less important edge', fontsize : 2},
               (2, 0) : {s : 'Very important edge', fontcolor : 'red'},
           }

    annotation_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text if only the annotation string is given.
        For a full list of available arguments see the matplotlib documentation.
    tables : dict node/edge : pandas dataframe
        Mapping of nodes and/or edges to pandas dataframes.
        The visibility of the tables that can toggled on or off by clicking on the corresponding node or edge.
    table_kwargs : dict
        Keyword arguments passed to matplotlib.pyplot.table.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`Graph`, :py:class:`EditableGraph`, :py:class:`InteractiveMultiGraph`, :py:class:`InteractiveArcDiagram`

    Notes
    -----
    You must retain a reference to the plot instance!
    Otherwise, the plot instance will be garbage collected after the initial draw
    and you won't be able to move the plot elements around.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from netgraph import InteractiveGraph
    >>> plt.ion()
    >>> plot_instance = InteractiveGraph(my_graph_obj)
    >>> plt.show()

    """

    def __init__(self, *args, **kwargs):
        DraggableGraphWithGridMode.__init__(self, *args, **kwargs)
        self._setup_emphasis()
        self._setup_annotations(*args, **kwargs)
        self._setup_table_annotations(*args, **kwargs)


    def _on_motion(self, event):
        DraggableGraphWithGridMode._on_motion(self, event)
        EmphasizeOnHoverGraph._on_motion(self, event)


    def _on_release(self, event):
        if self._currently_dragging is False:
            DraggableGraphWithGridMode._on_release(self, event)
            if self.artist_to_annotation:
                AnnotateOnClickGraph._on_release(self, event)
            if self.artist_to_table:
                TableOnClickGraph._on_release(self, event)
        else:
            DraggableGraphWithGridMode._on_release(self, event)
            if self.artist_to_annotation:
                self._redraw_annotations(event)


class NascentEdge(plt.Line2D):
    def __init__(self, source, origin):
        self.source = source
        self.origin = origin
        x0, y0 = origin
        super().__init__([x0, x0], [y0, y0], color='lightgray', linestyle='--')

    def _update(self, x1, y1):
        x0, y0 = self.origin
        super().set_data([[x0, x1], [y0, y1]])


class MutableGraph(InteractiveGraph):
    """Extends :py:class:`InteractiveGraph` to support the addition or removal of nodes and edges.

    - Double clicking on two nodes successively will create an edge between them.
    - Pressing 'insert' or '+' will add a new node to the graph.
    - Pressing 'delete' or '-' will remove selected nodes and edges.
    - Pressing '@' will reverse the direction of selected edges.

    When adding a new node, the properties of the last clicked node will be used to style the node artist.
    Ditto for edges. If no node or edge has been previously selected, the first created node or edge artist will be used.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    *args, **kwargs
        Parameters passed through to :py:class:`InteractiveGraph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`InteractiveGraph`, :py:class:`EditableGraph`, :py:class:`MutableMultiGraph`

    """

    def __init__(self, graph, *args, **kwargs):

        if is_order_zero(graph):
            # The graph is order-zero, i.e. it has no edges and no nodes.
            # We hence initialise with a single edge, which populates
            # - last_selected_node_properties
            # - last_selected_edge_properties
            # with the chosen parameters.
            # We then delete the edge and the two nodes and return the empty canvas.
            super().__init__([(0, 1)], *args, **kwargs)
            self._initialize_data_structures()
            self._delete_edge((0, 1))
            self._delete_node(0)
            self._delete_node(1)

        elif is_empty(graph):
            # The graph is empty, i.e. it has at least one node but no edges.
            nodes, _, _ = parse_graph(graph)
            if len(nodes) > 1:
                edge = (nodes[0], nodes[1])
                super().__init__([edge], nodes=nodes, *args, **kwargs)
                self._initialize_data_structures()
                self._delete_edge(edge)
            else: # single node
                node = nodes[0]
                dummy = 0 if node != 0 else 1
                edge = (node, dummy)
                super().__init__([edge], *args, **kwargs)
                self._initialize_data_structures()
                self._delete_edge(edge)
                self._delete_node(dummy)

        else:
            super().__init__(graph, *args, **kwargs)
            self._initialize_data_structures()

        # Ignore data limits and return full canvas.
        xmin, ymin = self.origin
        dx, dy = self.scale
        self.ax.axis([xmin, xmin+dx, ymin, ymin+dy])
        if self.autoscale_node_labels:
            self._rescale_node_labels()
        self.fig.canvas.draw()

        self.fig.canvas.mpl_connect('key_press_event', self._on_key_press)


    def _initialize_data_structures(self):
        self._reverse_node_artists = {artist : node for node, artist in self.node_artists.items()}
        self._reverse_edge_artists = {artist : edge for edge, artist in self.edge_artists.items()}
        self._last_selected_node_properties = self._extract_node_properties(next(iter(self.node_artists.values())))
        self._last_selected_node_type = type(next(iter(self.node_artists.values())))
        self._last_selected_edge_properties = self._extract_edge_properties(next(iter(self.edge_artists.values())))
        self._nascent_edge = None


    def _on_key_press(self, event):
        if event.key in ('insert', '+'):
            self._add_node(event)
        elif event.key in ('delete', '-'):
            self._delete_nodes()
            self._delete_edges()
        elif event.key == '@':
            self._reverse_edges()
        else:
            pass

        self.fig.canvas.draw_idle()


    def _on_press(self, event):
        # TODO : Potentially trigger this code on any node or edge selection;
        # clicking on a node or edge is just one of the ways to select them.
        # NOTE: However, clicking on an artist is guaranteed to be deliberate,
        # whereas other forms of selecting an artist are not.
        super()._on_press(event)

        if event.inaxes == self.ax:
            for artist in self._clickable_artists:
                if artist.contains(event)[0]:
                    self._extract_artist_properties(artist)
                    break

            if event.dblclick:
                self._add_or_remove_nascent_edge(event)
                self.fig.canvas.draw_idle()


    def _add_or_remove_nascent_edge(self, event):
        for node, artist in self.node_artists.items():
            if artist.contains(event)[0]:
                if self._nascent_edge:
                    # connect edge to target node
                    if (self._nascent_edge.source, node) not in self.edges:
                        self._add_edge(self._nascent_edge.source, node)
                        self.edge_layout.get()
                        self._update_edge_artists()
                    else:
                        print("Edge already exists!")
                    self._remove_nascent_edge()
                else:
                    self._nascent_edge = self._add_nascent_edge(node)
                break
        else:
            if self._nascent_edge:
                self._remove_nascent_edge()


    def _add_nascent_edge(self, node):
        nascent_edge = NascentEdge(node, self.node_positions[node])
        self.ax.add_artist(nascent_edge)
        return nascent_edge


    def _remove_nascent_edge(self):
        self._nascent_edge.remove()
        self._nascent_edge = None


    def _extract_artist_properties(self, artist):
        if isinstance(artist, NodeArtist):
            self._last_selected_node_properties = self._extract_node_properties(artist)
            self._last_selected_node_type = type(artist)
        elif isinstance(artist, EdgeArtist):
            self._last_selected_edge_properties = self._extract_edge_properties(artist)


    def _extract_node_properties(self, node_artist):
        properties = dict(
            size                 = node_artist.size,
            linewidth            = self._base_linewidth[node_artist],
            facecolor            = node_artist.get_facecolor(),
            edgecolor            = self._base_edgecolor[node_artist],
            alpha                = self._base_alpha[node_artist],
            zorder               = node_artist.get_zorder(),
        )
        if isinstance(node_artist, CircularNodeArtist):
            pass
        elif isinstance(node_artist, RegularPolygonNodeArtist):
            properties["orientation"] = node_artist.orientation
            properties["total_vertices"] = len(node_artist._path) - 1
        else: # NodeArtist
            properties["path"] = node_artist._path
            properties["orientation"] = node_artist.orientation
            properties["linewidth_correction"] = node_artist.linewidth_correction
        return properties


    def _extract_edge_properties(self, edge_artist):
        return dict(
            width       = edge_artist.width,
            facecolor   = edge_artist.get_facecolor(),
            alpha       = self._base_alpha[edge_artist],
            head_length = edge_artist.head_length,
            head_width  = edge_artist.head_width,
            edgecolor   = self._base_edgecolor[edge_artist],
            linewidth   = self._base_linewidth[edge_artist],
            head_offset = edge_artist.head_offset, # TODO: change to target node size
            tail_offset = edge_artist.tail_offset, # TODO: change to source node size
            curved      = edge_artist.curved,
            zorder      = edge_artist.get_zorder(),
        )


    def _on_motion(self, event):
        super()._on_motion(event)

        if event.inaxes == self.ax:
            if self._nascent_edge:
                self._nascent_edge._update(event.xdata, event.ydata)
                self.fig.canvas.draw_idle()


    def _add_node(self, event):
        if event.inaxes != self.ax:
            print('Position outside of axis limits! Cannot create node.')
            return

        # create node ID; use smallest unused int
        node = len(self.nodes)
        while node in self.node_positions.keys():
            node += 1

        # get position of cursor place node at cursor position
        pos = self._set_position_of_newly_created_node(event.xdata, event.ydata)

        # copy attributes of last selected artist
        node_properties = self._last_selected_node_properties
        node_type = self._last_selected_node_type

        artist = node_type(xy=pos, **node_properties)
        self.ax.add_patch(artist)
        self._expand_node_data_structures(node, artist, pos)

        return node


    def _expand_node_data_structures(self, node, artist, pos):
        # 0) MutableGraph
        self._reverse_node_artists[artist] = node
        # 1) InteractiveGraph
        # 2a) DraggableGraph
        self._draggable_artist_to_node[artist] = node
        # 2b) EmphasizeOnHoverGraph
        self.artist_to_key[artist] = node
        # 2c) AnnotateOnClickGraph
        # None
        # 3a) Graph
        # None
        # 3b) ClickableArtists, SelectableArtists, DraggableArtists
        self._clickable_artists.append(artist)
        self._selectable_artists.append(artist)
        self._draggable_artists.append(artist)
        self._base_linewidth[artist] = artist._lw_data
        self._base_edgecolor[artist] = artist.get_edgecolor()
        # 3c) EmphasizeOnHover
        self.emphasizeable_artists.append(artist)
        self._base_alpha[artist] = artist.get_alpha()
        # 3d) AnnotateOnClick
        # None
        # 4) BaseGraph
        self.nodes.append(node)
        self.node_positions[node] = pos
        self.node_artists[node] = artist
        # self.node_label_artists # TODO (potentially)
        # self.node_label_offset  # TODO (potentially)
        # 5) edge layout engine
        self.edge_layout.add_node(node, pos)


    def _set_position_of_newly_created_node(self, x, y):
        return (x, y)


    def _delete_nodes(self):
        # translate selected artists into nodes
        nodes = [self._reverse_node_artists[artist] for artist in self._selected_artists if isinstance(artist, NodeArtist)]

        # delete edges to and from selected nodes
        edges = [edge for edge in self.edges if ((edge[0] in nodes) or (edge[1] in nodes))]
        for edge in edges:
            self._delete_edge(edge)

        # delete nodes
        for node in nodes:
            self._delete_node(node)


    def _delete_node(self, node):
        # print(f"Deleting node {node}.")
        artist = self.node_artists[node]
        self._contract_node_data_structures(node, artist)
        artist.remove()


    def _contract_node_data_structures(self, node, artist):
        # 0) MutableGraph
        del self._reverse_node_artists[artist]
        # 1) InteractiveGraph
        # None
        # 2a) DraggableGraph
        del self._draggable_artist_to_node[artist]
        # 2b) EmphasizeOnHoverGraph
        del self.artist_to_key[artist]
        # None
        # 2c) AnnotateOnClickGraph
        if artist in self.annotated_artists:
            self._remove_annotation(artist)
        # 3a) Graph
        # None
        # 3b) ClickableArtists, SelectableArtists, DraggableArtists
        self._clickable_artists.remove(artist)
        self._selectable_artists.remove(artist)
        self._draggable_artists.remove(artist)
        if artist in self._selected_artists:
            self._selected_artists.remove(artist)
        del self._base_linewidth[artist]
        del self._base_edgecolor[artist]
        # 3c) EmphasizeOnHover
        self.emphasizeable_artists.remove(artist)
        if artist in self.deemphasized_artists:
            self.deemphasized_artists.remove(artist)
        del self._base_alpha[artist]
        # 3d) AnnotateOnClick
        if artist in self.artist_to_annotation:
            del self.artist_to_annotation[artist]
        # 4) BaseGraph
        self.nodes.remove(node)
        del self.node_positions[node]
        del self.node_artists[node]
        if hasattr(self, 'node_label_artists'):
            if node in self.node_label_artists:
                self.node_label_artists[node].remove()
                del self.node_label_artists[node]
        if hasattr(self, 'node_label_offset'):
            if node in self.node_label_offset:
                del self.node_label_offset[node]
        self.edge_layout.delete_node(node)



    def _add_edge(self, source, target, edge_properties=None):

        edge = (source, target)

        if not edge_properties:
            edge_properties = self._last_selected_edge_properties

        self.edge_layout.add_edge(edge)
        if source != target:
            edge_paths = self.edge_layout.approximate_nonloop_edge_paths([edge])
        else:
            edge_paths = self.edge_layout.approximate_selfloop_edge_paths([edge])
        path = edge_paths[edge]

        # create artist
        if (target, source) in self.edges:
            shape = 'right'
            self.edge_artists[(target, source)].shape = 'right'
            self.edge_artists[(target, source)]._update_path()
        else:
            shape = 'full'

        # create artist
        artist = EdgeArtist(midline=path, shape="full", **edge_properties)
        self.ax.add_patch(artist)

        # bookkeeping
        self._expand_edge_data_structures(edge, artist, path)

        return edge


    def _expand_edge_data_structures(self, edge, artist, path):
        # 0) MutableGraph
        self._reverse_edge_artists[artist] = edge
        # 1) InteractiveGraph
        # 2a) DraggableGraph
        # None
        # 2b) EmphasizeOnHoverGraph
        self.artist_to_key[artist] = edge
        # 2c) AnnotateOnClickGraph
        # None
        # 3a) Graph
        # None
        # 3b) ClickableArtists, SelectableArtists, DraggableArtists
        self._clickable_artists.append(artist)
        self._selectable_artists.append(artist)
        self._base_linewidth[artist] = artist._lw_data
        self._base_edgecolor[artist] = artist.get_edgecolor()
        # 3c) EmphasizeOnHover
        self.emphasizeable_artists.append(artist)
        self._base_alpha[artist] = artist.get_alpha()
        # 3d) AnnotateOnClick
        # None
        # 4) BaseGraph
        self.edges.append(edge)
        self.edge_artists[edge] = artist
        if hasattr(self, 'edge_label_artists'):
            self.draw_edge_labels(
                {edge : ""},
                self.edge_label_position,
                self.edge_label_rotate,
                self.edge_label_fontdict,
            )


    def _delete_edges(self):
        edges = [self._reverse_edge_artists[artist] for artist in self._selected_artists if isinstance(artist, EdgeArtist)]
        for edge in edges:
            self._delete_edge(edge)


    def _delete_edge(self, edge):
        source, target = edge
        if (target, source) in self.edges:
            self.edge_artists[(target, source)].shape = 'full'
            self.edge_artists[(target, source)]._update_path()

        artist = self.edge_artists[edge]
        self._contract_edge_data_structures(edge, artist)
        artist.remove()
        self.edge_layout.delete_edge(edge)


    def _contract_edge_data_structures(self, edge, artist):
        # 0) MutableGraph
        del self._reverse_edge_artists[artist]
        # 1) InteractiveGraph
        # None
        # 2a) DraggableGraph
        # None
        # 2b) EmphasizeOnHoverGraph
        del self.artist_to_key[artist]
        # 2c) AnnotateOnClickGraph
        if artist in self.annotated_artists:
            self._remove_annotation(artist)
        # 3a) Graph
        # None
        # 3b) ClickableArtists, SelectableArtists, DraggableArtists
        self._clickable_artists.remove(artist)
        self._selectable_artists.remove(artist)
        try:
            self._selected_artists.remove(artist)
        except ValueError:
            pass
        del self._base_linewidth[artist]
        del self._base_edgecolor[artist]
        # 3c) EmphasizeOnHover
        self.emphasizeable_artists.remove(artist)
        if artist in self.deemphasized_artists:
            self.deemphasized_artists.remove(artist)
        del self._base_alpha[artist]
        # 3d) AnnotateOnClick
        if artist in self.artist_to_annotation:
            del self.artist_to_annotation[artist]
        # 4) BaseGraph
        self.edges.remove(edge)
        del self.edge_artists[edge]
        if hasattr(self, 'edge_label_artists'):
            if edge in self.edge_label_artists:
                self.edge_label_artists[edge].remove()
                del self.edge_label_artists[edge]
        # TODO remove edge data


    def _reverse_edges(self):
        edges = [self._reverse_edge_artists[artist] for artist in self._selected_artists if isinstance(artist, EdgeArtist)]
        edge_properties = [self._extract_edge_properties(self.edge_artists[edge]) for edge in edges]

        # delete old edges;
        # note this step has to be completed before creating new edges,
        # as bi-directional edges can pose a problem otherwise
        for edge in edges:
            self._delete_edge(edge)

        for edge, properties in zip(edges, edge_properties):
            self._add_edge(edge[1], edge[0], properties)

        self.edge_layout.get()
        self._update_edge_artists()


class EditableGraph(MutableGraph):
    """Extends :py:class:`InteractiveGraph` to support adding, deleting, and editing graph elements interactively.

    a) Addition and removal of nodes and edges:

    - Double clicking on two nodes successively will create an edge between them.
    - Pressing 'insert' or '+' will add a new node to the graph.
    - Pressing 'delete' or '-' will remove selected nodes and edges.
    - Pressing '@' will reverse the direction of selected edges.

    b) Creation and editing of labels and annotations:

    - To create or edit a node or edge label, select the node (or edge) artist, press the 'enter' key, and type.
    - To create or edit an annotation, select the node (or edge) artist, press 'alt'+'enter', and type.
    - Terminate either action by pressing 'enter' or 'alt'+'enter' a second time.

    When adding a new node, the properties of the last clicked node will be used to style the node artist.
    Ditto for edges. If no node or edge has been previously selected the first created node or edge artist will be used.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    *args, **kwargs
        Parameters passed through to :py:class:`InteractiveGraph`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`InteractiveGraph`, :py:class:`EditableMultiGraph`

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not hasattr(self, 'node_label_artists'):
            self._initialize_empty_node_labels(kwargs)

        if not hasattr(self, 'edge_label_artists'):
            self._initialize_empty_edge_labels(kwargs)

        self._currently_writing_labels = False
        self._currently_writing_annotations = False

        # restore deprecated self.fig.canvas.manager.key_press method
        # https://github.com/matplotlib/matplotlib/issues/26713#issuecomment-1709938648
        self.fig.canvas.manager.key_press = key_press_handler


    def _initialize_empty_node_labels(self, kwargs):
        node_labels = {node : '' for node in self.nodes}
        self.node_label_fontdict = self._initialize_node_label_fontdict(
            kwargs.get('node_label_fontdict', None))
        self.node_label_offset, self._recompute_node_label_offsets =\
            self._initialize_node_label_offset(node_labels, kwargs.get('node_label_offset', (0., 0.)))
        if self._recompute_node_label_offsets:
            self._update_node_label_offsets()
        self.node_label_artists = dict()
        self.draw_node_labels(node_labels, self.node_label_fontdict)


    def _initialize_empty_edge_labels(self, kwargs):
        edge_labels = {edge : '' for edge in self.edges}
        self.edge_label_fontdict = self._initialize_edge_label_fontdict(kwargs.get('edge_label_fontdict'))
        self.edge_label_position = kwargs.get('edge_label_position', 0.5)
        self.edge_label_rotate = kwargs.get('edge_label_rotate', True)
        self.edge_label_artists = dict()
        self.draw_edge_labels(edge_labels, self.edge_label_position,
                              self.edge_label_rotate, self.edge_label_fontdict)


    def _on_key_press(self, event):
        if event.key == 'enter':
            if self._currently_writing_labels or self._currently_writing_annotations:
                self._terminate_writing()
            else:
                self._initiate_writing_labels()
        elif event.key == 'alt+enter':
            if self._currently_writing_annotations or self._currently_writing_labels:
                self._terminate_writing()
            else:
                self._initiate_writing_annotations()
        else:
            if self._currently_writing_labels:
                self._edit_labels(event.key)
            elif self._currently_writing_annotations:
                self._edit_annotations(event.key)
            else:
                super()._on_key_press(event)


    def _terminate_writing(self):
        self._currently_writing_labels = False
        self._currently_writing_annotations = False
        self.fig.canvas.manager.key_press_handler_id \
            = self.fig.canvas.mpl_connect('key_press_event', self.fig.canvas.manager.key_press)
        print('Finished writing.')


    def _initiate_writing_labels(self):
        self._currently_writing_labels = True
        self.fig.canvas.mpl_disconnect(self.fig.canvas.manager.key_press_handler_id)
        print('Initiated writing label(s).')


    def _initiate_writing_annotations(self):
        self._currently_writing_annotations = True
        self.fig.canvas.mpl_disconnect(self.fig.canvas.manager.key_press_handler_id)
        print('Initiated writing annotations(s).')


    def _edit_labels(self, key):
        for artist in self._selected_artists:
            if isinstance(artist, NodeArtist):
                self._edit_node_label(artist, key)
            elif isinstance(artist, EdgeArtist):
                self._edit_edge_label(artist, key)


    def _edit_node_label(self, artist, key):
        node = self.artist_to_key[artist]
        if node not in self.node_label_artists:
            # re-use a random offset to position node label;
            # we will improve the placement by updating all node label offsets
            self.node_label_offset[node] = next(iter(self.node_label_offset.values()))
            self._update_node_label_offsets()
            self.draw_node_labels({node : ''}, self.node_label_fontdict)

        self._edit_text_object(self.node_label_artists[node], key)


    def _edit_edge_label(self, artist, key):
        edge = self.artist_to_key[artist]
        if edge not in self.edge_label_artists:
            self.draw_edge_labels({edge : ''}, self.edge_label_position,
                                  self.edge_label_rotate, self.edge_label_fontdict)

        self._edit_text_object(self.edge_label_artists[edge], key)


    def _edit_annotations(self, key):
        for artist in self._selected_artists:
            if artist not in self.annotated_artists:
                if artist not in self.artist_to_annotation:
                    self.artist_to_annotation[artist] = ''
                self.annotated_artists.add(artist)
                placement = self._get_annotation_placement(artist)
                self._add_annotation(artist, *placement)

            self._edit_text_object(self.artist_to_text_object[artist], key)
            self.artist_to_annotation[artist] = self.artist_to_text_object[artist].get_text()


    def _edit_text_object(self, text_object, key):
        if len(key) == 1:
            text_object.set_text(text_object.get_text() + key)
        elif key == 'backspace':
            text_object.set_text(text_object.get_text()[:-1])
        self.fig.canvas.draw_idle()
